# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The ExpressionManager is used to create expressions.

All objects are memoized so that two syntactically equivalent expressions
are represented by the same object.
"""

import unified_planning
import unified_planning.model.types
from unified_planning.model.operators import OperatorKind
from unified_planning.exceptions import UPTypeError, UPExpressionDefinitionError
from fractions import Fraction
from typing import Iterable, List, Union, Dict, Tuple

Expression = Union['unified_planning.model.fnode.FNode', 'unified_planning.model.fluent.Fluent', 'unified_planning.model.object.Object', 'unified_planning.model.parameter.Parameter', 'unified_planning.model.variable.Variable', bool, int, float, Fraction]
BoolExpression = Union['unified_planning.model.fnode.FNode', 'unified_planning.model.fluent.Fluent', 'unified_planning.model.parameter.Parameter', bool]

class ExpressionManager(object):
    """ExpressionManager is responsible for the creation of all expressions."""

    def __init__(self, env: 'unified_planning.environment.Environment'):
        self.env = env
        self.expressions: Dict['unified_planning.model.fnode.FNodeContent', 'unified_planning.model.fnode.FNode'] = {}
        self._next_free_id = 1

        self.true_expression = self.create_node(node_type=OperatorKind.BOOL_CONSTANT,
                                                args=tuple(),
                                                payload=True)
        self.false_expression = self.create_node(node_type=OperatorKind.BOOL_CONSTANT,
                                                 args=tuple(),
                                                 payload=False)
        return

    def _polymorph_args_to_tuple(self, *args: Union[Expression, Iterable[Expression]]) -> Tuple[Expression, ...]:
        """ Helper function to return a tuple of arguments from args.
        This function is used to allow N-ary operators to express their arguments
        both as a list of arguments or as a tuple of arguments: e.g.,
           And([a,b,c]) and And(a,b,c)
        are both valid, and they are converted into a tuple (a,b,c) """

        res = []
        for p in args:
            if isinstance(p, Iterable):
                res.extend(list(p))
            else:
                res.append(p)
        return tuple(res)

    def auto_promote(self, *args: Union[Expression, Iterable[Expression]]) -> List['unified_planning.model.fnode.FNode']:
        tuple_args = self._polymorph_args_to_tuple(*args)
        res = []
        for e in tuple_args:
            if isinstance(e, unified_planning.model.fluent.Fluent):
                res.append(self.FluentExp(e))
            elif isinstance(e, unified_planning.model.parameter.Parameter):
                res.append(self.ParameterExp(e))
            elif isinstance(e, unified_planning.model.variable.Variable):
                res.append(self.VariableExp(e))
            elif isinstance(e, unified_planning.model.object.Object):
                res.append(self.ObjectExp(e))
            elif isinstance(e, bool):
                res.append(self.Bool(e))
            elif isinstance(e, int):
                res.append(self.Int(e))
            elif isinstance(e, float):
                res.append(self.Real(Fraction(e)))
            elif isinstance(e, Fraction):
                res.append(self.Real(e))
            else:
                res.append(e)
        return res

    def create_node(self, node_type: OperatorKind, args: Iterable['unified_planning.model.fnode.FNode'],
                    payload: Union['unified_planning.model.fluent.Fluent', 'unified_planning.model.object.Object', 'unified_planning.model.parameter.Parameter', 'unified_planning.model.variable.Variable', bool, int, Fraction, Tuple['unified_planning.model.variable.Variable', ...]] = None) ->'unified_planning.model.fnode.FNode':
        content = unified_planning.model.fnode.FNodeContent(node_type, args, payload)
        if content in self.expressions:
            return self.expressions[content]
        else:
            n =unified_planning.model.fnode.FNode(content, self._next_free_id)
            self._next_free_id += 1
            self.expressions[content] = n
            self.env.type_checker.get_type(n)
            return n

    def And(self, *args: Union[BoolExpression, Iterable[BoolExpression]]) -> 'unified_planning.model.fnode.FNode':
        """ Returns a conjunction of terms.
        This function has polimorphic n-arguments:
          - And(a,b,c)
          - And([a,b,c])
        Restriction: Arguments must be boolean
        """
        tuple_args = tuple(self.auto_promote(*args))

        if len(tuple_args) == 0:
            return self.TRUE()
        elif len(tuple_args) == 1:
            return tuple_args[0]
        else:
            return self.create_node(node_type=OperatorKind.AND,
                                    args=tuple_args)

    def Or(self, *args: Union[BoolExpression, Iterable[BoolExpression]]) ->'unified_planning.model.fnode.FNode':
        """ Returns an disjunction of terms.
        This function has polimorphic n-arguments:
          - Or(a,b,c)
          - Or([a,b,c])
        Restriction: Arguments must be boolean
        """
        tuple_args = tuple(self.auto_promote(*args))

        if len(tuple_args) == 0:
            return self.FALSE()
        elif len(tuple_args) == 1:
            return tuple_args[0]
        else:
            return self.create_node(node_type=OperatorKind.OR,
                                    args=tuple_args)

    def Not(self, expression: BoolExpression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form:
                not expression
        Restriction: expression must be of boolean type
        """
        expression, = self.auto_promote(expression)
        if expression.is_not():
            return expression.arg(0)
        return self.create_node(node_type=OperatorKind.NOT, args=(expression,))

    def Implies(self, left: BoolExpression, right: BoolExpression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form:
            left -> right
        Restriction: Left and Right must be of boolean type
        """
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.IMPLIES, args=(left, right))

    def Iff(self, left: BoolExpression, right: BoolExpression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form:
            left <-> right
        Restriction: Left and Right must be of boolean type
        """
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.IFF, args=(left, right))

    def Exists(self, expression: BoolExpression, *vars: 'unified_planning.model.variable.Variable') ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form:
            Exists (var[0]... var[n]) | expression
        Restriction: expression must be of boolean type and
                    vars must be of 'unified_planning.Variable' type
        """
        expressions = tuple(self.auto_promote(expression))
        if len(vars) == 0:
            raise UPExpressionDefinitionError(f"Exists of expression: {str(expression)} must be created with at least one variable, otherwise it is not needed.")
        for v in vars:
            if not isinstance(v, unified_planning.model.variable.Variable):
                raise UPTypeError("Expecting 'unified_planning.Variable', got %s", type(v))
        return self.create_node(node_type=OperatorKind.EXISTS, args=expressions, payload=vars)

    def Forall(self, expression: BoolExpression, *vars: 'unified_planning.model.variable.Variable') ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form:
            Forall (var[0]... var[n]) | expression
        Restriction: expression must be of boolean type and
                    vars must be of 'unified_planning.Variable' type
        """
        expressions = tuple(self.auto_promote(expression))
        if len(vars) == 0:
            raise UPExpressionDefinitionError(f"Forall of expression: {str(expression)} must be created with at least one variable, otherwise it is not needed.")
        for v in vars:
            if not isinstance(v, unified_planning.model.variable.Variable):
                raise UPTypeError("Expecting 'unified_planning.Variable', got %s", type(v))
        return self.create_node(node_type=OperatorKind.FORALL, args=expressions, payload=vars)

    def FluentExp(self, fluent: 'unified_planning.model.fluent.Fluent', params: Tuple[Expression, ...] = tuple()) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression for the given fluent and parameters.
        Restriction: parameters type must be compatible with the fluent signature
        """
        assert fluent.arity == len(params)
        params_exp = self.auto_promote(*params)
        return self.create_node(node_type=OperatorKind.FLUENT_EXP, args=tuple(params_exp), payload=fluent)

    def ParameterExp(self, param: 'unified_planning.model.parameter.Parameter') ->'unified_planning.model.fnode.FNode':
        """Returns an expression for the given action parameter."""
        return self.create_node(node_type=OperatorKind.PARAM_EXP, args=tuple(), payload=param)

    def VariableExp(self, var: 'unified_planning.model.variable.Variable') ->'unified_planning.model.fnode.FNode':
        """Returns an expression for the given variable."""
        return self.create_node(node_type=OperatorKind.VARIABLE_EXP, args=tuple(), payload=var)

    def ObjectExp(self, obj: 'unified_planning.model.object.Object') ->'unified_planning.model.fnode.FNode':
        """Returns an expression for the given object."""
        return self.create_node(node_type=OperatorKind.OBJECT_EXP, args=tuple(), payload=obj)

    def TRUE(self) ->'unified_planning.model.fnode.FNode':
        """Return the boolean constant True."""
        return self.true_expression

    def FALSE(self) ->'unified_planning.model.fnode.FNode':
        """Return the boolean constant False."""
        return self.false_expression

    def Bool(self, value: bool) ->'unified_planning.model.fnode.FNode':
        """Return a boolean constant."""
        if type(value) != bool:
            raise UPTypeError("Expecting bool, got %s" % type(value))

        if value:
            return self.true_expression
        else:
            return self.false_expression

    def Int(self, value: int) ->'unified_planning.model.fnode.FNode':
        """Return an int constant."""
        if type(value) != int:
            raise UPTypeError("Expecting int, got %s" % type(value))
        return self.create_node(node_type=OperatorKind.INT_CONSTANT, args=tuple(), payload=value)

    def Real(self, value: Fraction) ->'unified_planning.model.fnode.FNode':
        """Return a real constant."""
        if type(value) != Fraction:
            raise UPTypeError("Expecting Fraction, got %s" % type(value))
        return self.create_node(node_type=OperatorKind.REAL_CONSTANT, args=tuple(), payload=value)

    def Plus(self, *args: Union[Expression, Iterable[Expression]]) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form:
            args[0] + ... + args[n]
        """
        tuple_args = tuple(self.auto_promote(*args))

        if len(tuple_args) == 0:
            return self.Int(0)
        elif len(tuple_args) == 1:
            return tuple_args[0]
        else:
            return self.create_node(node_type=OperatorKind.PLUS,
                                    args=tuple_args)

    def Minus(self, left: Expression, right: Expression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form: left - right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.MINUS, args=(left, right))

    def Times(self, *args: Union[Expression, Iterable[Expression]]) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form:
            args[0] * ... * args[n]
        """
        tuple_args = tuple(self.auto_promote(*args))

        if len(tuple_args) == 0:
            return self.Int(1)
        elif len(tuple_args) == 1:
            return tuple_args[0]
        else:
            return self.create_node(node_type=OperatorKind.TIMES,
                                    args=tuple_args)

    def Div(self, left: Expression, right: Expression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form: left / right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.DIV, args=(left, right))

    def LE(self, left: Expression, right: Expression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form: left <= right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.LE, args=(left, right))

    def GE(self, left: Expression, right: Expression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form: left >= right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.LE, args=(right, left))

    def LT(self, left: Expression, right: Expression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form: left < right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.LT, args=(left, right))

    def GT(self, left: Expression, right: Expression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form: left > right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.LT, args=(right, left))

    def Equals(self, left: Expression, right: Expression) ->'unified_planning.model.fnode.FNode':
        """ Creates an expression of the form: left == right."""
        left, right = self.auto_promote(left, right)
        return self.create_node(node_type=OperatorKind.EQUALS, args=(left, right))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial

import sys
if sys.version_info >= (3, 3):
    from collections.abc import Iterable
else:
    from collections import Iterable


from unified_planning.model import FNode, OperatorKind

# NodeType to Function Name
def nt_to_fun(o: OperatorKind) -> str:
    """Returns the name of the walk function for the given nodetype."""
    return "walk_%s" % (str(o).replace('OperatorKind.', '')).lower()

class handles(object):
    """Decorator for walker functions.
    Use it by specifying the nodetypes that need to be handled by the
    given function. It is possible to use groupd (e.g., OperatorKind.RELATIONS)
    directly. ::
      @handles(OperatorKind.NODE, ...)
      def walk_special(...):
         ...
    """
    def __init__(self, *nodetypes):
        if len(nodetypes) == 1 and isinstance(nodetypes[0], Iterable):
            nodetypes = nodetypes[0]
        self.nodetypes = list(nodetypes)

    def __call__(self, func):
        nodetypes = self.nodetypes
        if hasattr(func, "nodetypes"):
            nodetypes = func.nodetypes + nodetypes
        func.nodetypes = nodetypes
        return func

class MetaNodeTypeHandler(type):
    """Metaclass used to intepret the nodehandler decorator. """
    def __new__(cls, name, bases, dct):
        obj = type.__new__(cls, name, bases, dct)
        for k,v in dct.items():
            if hasattr(v, "nodetypes"):
                obj.set_handler(v, *v.nodetypes)
        return obj


class Walker(object, metaclass=MetaNodeTypeHandler):
    """Base Abstract Walker class.
    Do not subclass directly, use DagWalker or TreeWalker, instead.
    """

    def __init__(self):
        self.functions = {}
        for o in iter(OperatorKind):
            try:
                # getattr will raise an AttributeError exception if a
                # method does not exist
                self.functions[o] = getattr(self, nt_to_fun(o))
            except AttributeError:
                self.functions[o] = self.walk_error

    def set_function(self, function, *node_types):
        """Overrides the default walking function for each of the specified
        node_types with the given function
        """
        from warnings import warn
        warn("Instance-based walkers (<=0.6.0) walkers are deprecated. "
             "You should use new-style/class based walkers", stacklevel=2)
        for nt in node_types:
            self.functions[nt] = function

    @classmethod
    def set_handler(cls, function, *node_types):
        """Associate in cls the given function to the given node_types."""
        for nt in node_types:
            setattr(cls, nt_to_fun(nt), function)

    @classmethod
    def super(cls, self, expression: FNode, *args, **kwargs):
        """Call the correct walk_* function of cls for the given expression."""
        f = getattr(cls, nt_to_fun(expression.node_type))
        return f(self, expression, *args, **kwargs)

    @handles(OperatorKind)
    def walk_error(self, expression: FNode, **kwargs):
        raise NotImplementedError

# EOC Walker



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the conditional effects remover class."""


import unified_planning
from unified_planning.exceptions import UPProblemDefinitionError
from unified_planning.transformers.transformer import Transformer
from typing import Iterable, List, Dict, Tuple, Union, Optional
from itertools import chain, combinations


class ConditionalEffectsRemover(Transformer):
    '''Conditional effect remover class:
    this class requires a problem and offers the capability
    to transform a conditional problem into an unconditional one.

    This is done by substituting every conditional action with different
    actions representing every possible branch of the original action.

    Also the conditional timed_effects are removed maintaining the same
    semanthics. When this is not possible, an exception is raised.'''
    def __init__(self, problem: 'unified_planning.model.Problem', name: str = 'cerm'):
        Transformer.__init__(self, problem, name)
        #Represents the map from the new action to the old action
        self._new_to_old: Dict['unified_planning.model.Action', 'unified_planning.model.Action'] = {}
        #represents a mapping from the action of the original problem to action of the new one.
        self._old_to_new: Dict['unified_planning.model.Action', List['unified_planning.model.Action']] = {}

    def powerset(self, iterable: Iterable) -> Iterable:
        "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
        s = list(iterable)
        return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

    def get_rewritten_problem(self) -> 'unified_planning.model.Problem':
        '''Creates a problem that is a copy of the original problem
        but every conditional action or effect are removed.'''
        if self._new_problem is not None:
            return self._new_problem
        #NOTE that a different environment might be needed when multi-threading
        self._new_problem = self._problem.clone()
        self._new_problem.name = f'{self._name}_{self._problem.name}'
        self._new_problem.clear_timed_effects()
        for t, el in self._problem.timed_effects.items():
            for e in el:
                if e.is_conditional():
                    f, v = e.fluent.fluent(), e.value
                    if not f.type.is_bool_type():
                        raise UPProblemDefinitionError(f'The condition of effect: {e}\ncould not be removed without changing the problem.')
                    else:
                        em = self._env.expression_manager
                        c = e.condition
                        nv = self._simplifier.simplify(em.Or(em.And(c, v), em.And(em.Not(c), f)))
                        self._new_problem.add_timed_effect(t, e.fluent, nv)
                else:
                    self._new_problem._add_effect_instance(t, e.clone())
        self._new_problem.clear_actions()
        for ua in self._problem.unconditional_actions:
            new_uncond_action = ua.clone()
            self._new_problem.add_action(new_uncond_action)
            self._new_to_old[new_uncond_action] = ua
            self._map_old_to_new_action(ua, new_uncond_action)
        for action in self._problem.conditional_actions:
            if isinstance(action, unified_planning.model.InstantaneousAction):
                cond_effects = action.conditional_effects
                for p in self.powerset(range(len(cond_effects))):
                    new_action = action.clone()
                    new_action.name = self.get_fresh_name(action.name)
                    new_action.clear_effects()
                    for e in action.unconditional_effects:
                        new_action._add_effect_instance(e.clone())
                    for i, e in enumerate(cond_effects):
                        if i in p:
                            # positive precondition
                            new_action.add_precondition(e.condition)
                            ne = unified_planning.model.Effect(e.fluent, e.value, self._env.expression_manager.TRUE(), e.kind)
                            new_action._add_effect_instance(ne)
                        else:
                            #negative precondition
                            new_action.add_precondition(self._env.expression_manager.Not(e.condition))
                    #new action is created, then is checked if it has any impact and if it can be simplified
                    if len(new_action.effects) > 0:
                        action_is_feasible, simplified_preconditions = self._check_and_simplify_preconditions(new_action)
                        if action_is_feasible:
                            new_action._set_preconditions(simplified_preconditions)
                            self._new_to_old[new_action] = action
                            self._map_old_to_new_action(action, new_action)
                            self._new_problem.add_action(new_action)
            elif isinstance(action, unified_planning.model.DurativeAction):
                timing_cond_effects: Dict['unified_planning.model.Timing', List['unified_planning.model.Effect']] = action.conditional_effects
                cond_effects_timing: List[Tuple['unified_planning.model.Effect', 'unified_planning.model.Timing']] = [(e, t) for t, el in timing_cond_effects.items() for e in el]
                for p in self.powerset(range(len(cond_effects_timing))):
                    new_action = action.clone()
                    new_action.name = self.get_fresh_name(action.name)
                    new_action.clear_effects()
                    for t, el in action.unconditional_effects.items():
                        for e in el:
                            new_action._add_effect_instance(t, e.clone())
                    for i, (e, t) in enumerate(cond_effects_timing):
                        if i in p:
                            # positive precondition
                            new_action.add_condition(t, e.condition)
                            ne = unified_planning.model.Effect(e.fluent, e.value, self._env.expression_manager.TRUE(), e.kind)
                            new_action._add_effect_instance(t, ne)
                        else:
                            #negative precondition
                            new_action.add_condition(t, self._env.expression_manager.Not(e.condition))
                    #new action is created, then is checked if it has any impact and if it can be simplified
                    if len(new_action.effects) > 0:
                        action_is_feasible, simplified_conditions = self._check_and_simplify_conditions(new_action)
                        if action_is_feasible:
                            new_action.clear_conditions()
                            for interval, c in simplified_conditions:
                                new_action.add_condition(interval, c)
                            self._new_to_old[new_action] = action
                            self._map_old_to_new_action(action, new_action)
                            self._new_problem.add_action(new_action)
            else:
                raise NotImplementedError
        return self._new_problem

    def _map_old_to_new_action(self, old_action, new_action):
        if old_action in self._old_to_new:
            self._old_to_new[old_action].append(new_action)
        else:
            self._old_to_new[old_action] = [new_action]

    def get_original_action(self, action: 'unified_planning.model.Action') -> 'unified_planning.model.Action':
        '''After the method get_rewritten_problem is called, this function maps
        the actions of the transformed problem into the actions of the original problem.'''
        return self._new_to_old[action]

    def get_transformed_actions(self, action: 'unified_planning.model.Action') -> List['unified_planning.model.Action']:
        '''After the method get_rewritten_problem is called, this function maps
        the actions of the original problem into the actions of the transformed problem.'''
        return self._old_to_new[action]



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
'''This module defines the different remover classes.'''


import unified_planning
from unified_planning.model import FNode, Timing, TimeInterval, Action, InstantaneousAction, DurativeAction, Problem
from unified_planning.plan import SequentialPlan, TimeTriggeredPlan, ActionInstance, Plan
from typing import Dict, Iterable, List, Optional, OrderedDict, Tuple, Union


class Transformer:
    '''Represents a generic Transformer with all the needed methods shared among them.'''
    def __init__(self, problem: Problem, name: str):
        self._name = name
        self._problem = problem
        self._env = problem.env
        self._new_problem: Optional[Problem] = None
        self._simplifier = unified_planning.walkers.Simplifier(self._env)

    def get_original_action(self, action: Action) -> Action:
        '''After the method get_rewritten_problem is called, this function should
         map the actions of the transformed problem into the actions of the original problem.'''
        raise NotImplementedError

    def get_transformed_actions(self, action: Action) -> List[Action]:
        '''After the method get_rewritten_problem is called, this function should
         map the actions of the original problem into the actions of the transformed problem.'''
        raise NotImplementedError

    def get_rewritten_problem(self) -> Problem:
        '''This function should rewrite the problem according to the Transformer specifics.
        This should also include the data structure to support the methods get_original_action
        and get_transformed_actions'''
        raise NotImplementedError

    def rewrite_back_plan(self, plan: Plan) -> Plan:
        '''Takes the sequential plan of the problem (created with
        the method "self.get_rewritten_problem()" and translates the plan back
        to be a plan of the original problem.

        NOTE:
        This method MUST be rewritten if the specific Transformer extension changes
        the action's parameters!'''
        if isinstance(plan, SequentialPlan):
            new_actions: List[ActionInstance] = plan.actions
            old_actions: List[ActionInstance] = []
            for ai in new_actions:
                old_actions.append(ActionInstance(self.get_original_action(ai.action), ai.actual_parameters))
            return SequentialPlan(old_actions)
        elif isinstance(plan, TimeTriggeredPlan):
            s_new_actions_d = plan.actions
            s_old_actions_d = []
            for s, ai, d in s_new_actions_d:
                s_old_actions_d.append((s, ActionInstance(self.get_original_action(ai.action), ai.actual_parameters), d))
            return TimeTriggeredPlan(s_old_actions_d)
        raise NotImplementedError

    def _check_and_simplify_conditions(self, action: DurativeAction, simplify_constants: bool = False) -> Tuple[bool, List[Tuple[TimeInterval, FNode]]]:
        '''Simplifies conditions and if it is False (a contraddiction)
        returns False, otherwise returns True.
        If the simplification is True (a tautology) removes all conditions at the given timing.
        If the simplification is still an AND rewrites back every "arg" of the AND
        in the conditions
        If the simplification is not an AND sets the simplification as the only
        condition at the given timing.
        Then, the new conditions are returned as a List[Tuple[Timing, FNode]] and the user can
        decide how to use the new conditions.'''
        #new action conditions
        nac: List[Tuple[TimeInterval, FNode]] = []
        # t = timing, lc = list condition
        for i, lc in action.conditions.items():
            #conditions (as an And FNode)
            c = self._env.expression_manager.And(lc)
            #conditions simplified
            if simplify_constants:
                cs = self._simplifier.simplify(c, self._problem)
            else:
                cs = self._simplifier.simplify(c)
            if cs.is_bool_constant():
                if not cs.bool_constant_value():
                    return (False, [],)
            else:
                if cs.is_and():
                    for new_cond in cs.args:
                        nac.append((i, new_cond))
                else:
                    nac.append((i, cs))
        return (True, nac)

    def _check_and_simplify_preconditions(self, action: InstantaneousAction, simplify_constants: bool = False) -> Tuple[bool, List[FNode]]:
        '''Simplifies preconditions and if it is False (a contraddiction)
        returns False, otherwise returns True.
        If the simplification is True (a tautology) removes all preconditions.
        If the simplification is still an AND rewrites back every "arg" of the AND
        in the preconditions
        If the simplification is not an AND sets the simplification as the only
        precondition.
        Then, the new preconditions are returned as a List[FNode] and the user can
        decide how to use the new preconditions.'''
        #action preconditions
        ap = action.preconditions
        if len(ap) == 0:
            return (True, [])
        #preconditions (as an And FNode)
        p = self._env.expression_manager.And(ap)
        #preconditions simplified
        if simplify_constants:
            ps = self._simplifier.simplify(p, self._problem)
        else:
            ps = self._simplifier.simplify(p)
        #new action preconditions
        nap: List[FNode] = []
        if ps.is_bool_constant():
            if not ps.bool_constant_value():
                return (False, [])
        else:
            if ps.is_and():
                nap.extend(ps.args)
            else:
                nap.append(ps)
        action._set_preconditions(nap)
        return (True, nap)

    def get_fresh_name(self, original_name: str, parameters_names: Iterable[str] = []) -> str:
        '''To use this method, the new problem returned by the transformer must be stored in the field
        self._new_problem!
        This method returns a fresh name for the problem, given the name of the transformer and a name in input.

        NOTE: The field parameters_names is there just for possible extensions.'''
        assert self._new_problem is not None
        if parameters_names != []:
            raise NotImplementedError
        count = 0
        while(True):
            new_name = f'{self._name}_{original_name}_{str(count)}'
            if self._problem.has_name(new_name) or self._new_problem.has_name(new_name):
                count += 1
            else:
                return new_name



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import unified_planning
from unified_planning.shortcuts import *
from unified_planning.exceptions import UPProblemDefinitionError
from unified_planning.model import GlobalStartTiming
from unified_planning.model.problem_kind import classical_kind, full_classical_kind, basic_temporal_kind
from unified_planning.test import TestCase, main
from unified_planning.test import skipIfNoPlanValidatorForProblemKind, skipIfNoOneshotPlannerForProblemKind, skipIfSolverNotAvailable
from unified_planning.test.examples import get_example_problems
from unified_planning.transformers import ConditionalEffectsRemover


class TestConditionalEffectsRemover(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    def test_basic_conditional(self):
        problem = self.problems['basic_conditional'].problem
        cer = ConditionalEffectsRemover(problem)
        unconditional_problem = cer.get_rewritten_problem()
        u_actions = unconditional_problem.actions
        a_x = problem.action("a_x")
        a_x_new_list = cer.get_transformed_actions(a_x)
        self.assertEqual(len(a_x_new_list), 1)
        new_action = unconditional_problem.action(a_x_new_list[0].name)
        y = FluentExp(problem.fluent("y"))
        a_x_e = a_x.effects
        new_action_e = new_action.effects

        self.assertEqual(len(u_actions), 2)
        self.assertEqual(problem.action("a_y"), unconditional_problem.action("a_y"))
        self.assertTrue(a_x.is_conditional())
        self.assertFalse(unconditional_problem.has_action("a_x"))
        self.assertFalse(new_action.is_conditional())
        self.assertIn(y, new_action.preconditions)
        self.assertNotIn(y, a_x.preconditions)
        for e, ue in zip(a_x_e, new_action_e):
            self.assertEqual(e.fluent, ue.fluent)
            self.assertEqual(e.value, ue.value)
            self.assertFalse(ue.is_conditional())

        self.assertTrue(problem.kind.has_conditional_effects())
        self.assertFalse(unconditional_problem.kind.has_conditional_effects())
        for a in unconditional_problem.actions:
            self.assertFalse(a.is_conditional())

        with OneshotPlanner(problem_kind=unconditional_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uncond_plan = planner.solve(unconditional_problem).plan
            new_plan = cer.rewrite_back_plan(uncond_plan)
            with PlanValidator(problem_kind=problem.kind) as pv:
                self.assertTrue(pv.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    def test_complex_conditional(self):
        problem = self.problems['complex_conditional'].problem
        cer = ConditionalEffectsRemover(problem)
        unconditional_problem = cer.get_rewritten_problem()
        unconditional_problem_2 = cer.get_rewritten_problem()
        self.assertEqual(unconditional_problem, unconditional_problem_2)

        with OneshotPlanner(problem_kind=unconditional_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uncond_plan = planner.solve(unconditional_problem).plan
            new_plan = cer.rewrite_back_plan(uncond_plan)
            with PlanValidator(problem_kind=problem.kind) as pv:
                self.assertTrue(pv.validate(problem, new_plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_temporal_kind))
    def test_temporal_conditional(self):
        problem = self.problems['temporal_conditional'].problem

        cer = ConditionalEffectsRemover(problem)
        unconditional_problem = cer.get_rewritten_problem()

        with OneshotPlanner(problem_kind=unconditional_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            uncond_plan = planner.solve(unconditional_problem).plan
            new_plan = cer.rewrite_back_plan(uncond_plan)
            for (s, a, d), (s_1, a_1, d_1) in zip(new_plan.actions, uncond_plan.actions):
                self.assertEqual(s, s_1)
                self.assertEqual(d, d_1)
                self.assertIn(a.action, problem.actions)

    def test_ad_hoc_1(self):
        ct = GlobalStartTiming(2)
        x = unified_planning.model.Fluent('x')
        y = unified_planning.model.Fluent('y')
        problem = unified_planning.model.Problem('ad_hoc_1')
        problem.add_fluent(x, default_initial_value=True)
        problem.add_fluent(y, default_initial_value=True)
        problem.add_timed_effect(ct, y, Not(x), x)
        problem.add_goal(Not(y))
        cer = ConditionalEffectsRemover(problem)
        uncond_problem = cer.get_rewritten_problem()
        eff = uncond_problem.timed_effects[ct][0]
        self.assertFalse(eff.is_conditional())
        self.assertEqual(FluentExp(y), eff.fluent)
        self.assertEqual(And(Not(x), y), eff.value)

    def test_mockup_1(self):
        ct = GlobalStartTiming(2)
        x = unified_planning.model.Fluent('x', IntType())
        y = unified_planning.model.Fluent('y')
        problem = unified_planning.model.Problem('mockup_1')
        problem.add_fluent(x, default_initial_value=0)
        problem.add_fluent(y, default_initial_value=True)
        problem.add_timed_effect(ct, y, False)
        problem.add_timed_effect(ct, x, 5, y)
        cer = ConditionalEffectsRemover(problem)
        with self.assertRaises(UPProblemDefinitionError) as e:
            cer.get_rewritten_problem()
        self.assertIn('The condition of effect: if y then x := 5\ncould not be removed without changing the problem.', str(e.exception))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import os
import tempfile
from typing import cast
import pytest
import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.io import PDDLWriter, PDDLReader
from unified_planning.test.examples import get_example_problems
from unified_planning.model.types import _UserType


FILE_PATH = os.path.dirname(os.path.abspath(__file__))
PDDL_DOMAINS_PATH = os.path.join(FILE_PATH, 'pddl')


class TestPddlIO(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_basic_writer(self):
        problem = self.problems['basic'].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn('(:requirements :strips :negative-preconditions)', pddl_domain)
        self.assertIn('(:predicates (x))', pddl_domain)
        self.assertIn('(:action a', pddl_domain)
        self.assertIn(':parameters ()', pddl_domain)
        self.assertIn(':precondition (and (not (x)))', pddl_domain)
        self.assertIn(':effect (and (x))', pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn('(:domain basic-domain)', pddl_problem)
        self.assertIn('(:init)', pddl_problem)
        self.assertIn('(:goal (and (x)))', pddl_problem)


    def test_basic_conditional_writer(self):
        problem = self.problems['basic_conditional'].problem

        self.assertTrue(problem.action('a_x').is_conditional())
        self.assertFalse(problem.action('a_y').is_conditional())

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn('(:requirements :strips :negative-preconditions :conditional-effects)', pddl_domain)
        self.assertIn('(:predicates (x) (y))', pddl_domain)
        self.assertIn('(:action a_x', pddl_domain)
        self.assertIn(':parameters ()', pddl_domain)
        self.assertIn(':precondition (and (not (x)))', pddl_domain)
        self.assertIn(':effect (and (when (y) (x)))', pddl_domain)
        self.assertIn('(:action a_y', pddl_domain)
        self.assertIn(':parameters ()', pddl_domain)
        self.assertIn(':precondition (and (not (y)))', pddl_domain)
        self.assertIn(':effect (and (y))', pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn('(:domain basic_conditional-domain)', pddl_problem)
        self.assertIn('(:init)', pddl_problem)
        self.assertIn('(:goal (and (x)))', pddl_problem)

    def test_basic_exists_writer(self):
        problem = self.problems['basic_exists'].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn('(:requirements :strips :typing :existential-preconditions)', pddl_domain)
        self.assertIn('(:predicates (x) (y ?semaphore - Semaphore))', pddl_domain)
        self.assertIn('(:action a', pddl_domain)
        self.assertIn(':parameters ()', pddl_domain)
        self.assertIn(':precondition (and (exists (?s - Semaphore)\n (y ?s)))', pddl_domain)
        self.assertIn(':effect (and (x))', pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn('(:domain basic_exists-domain)', pddl_problem)
        self.assertIn('(:objects \n   o1 o2 - Semaphore\n )', pddl_problem)
        self.assertIn('(:init (y o1))', pddl_problem)
        self.assertIn('(:goal (and (x)))', pddl_problem)

    def test_robot_writer(self):
        problem = self.problems['robot'].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn('(:requirements :strips :typing :negative-preconditions :equality :numeric-fluents)', pddl_domain)
        self.assertIn('(:types Location)', pddl_domain)
        self.assertIn('(:predicates (robot_at ?position - Location))', pddl_domain)
        self.assertIn('(:functions (battery_charge))', pddl_domain)
        self.assertIn('(:action move', pddl_domain)
        self.assertIn(':parameters ( ?l_from - Location ?l_to - Location)', pddl_domain)
        self.assertIn(':precondition (and (<= 10 (battery_charge)) (not (= ?l_from ?l_to)) (robot_at ?l_from) (not (robot_at ?l_to)))', pddl_domain)
        self.assertIn(':effect (and (not (robot_at ?l_from)) (robot_at ?l_to) (assign (battery_charge) (- (battery_charge) 10)))', pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn('(:domain robot-domain)', pddl_problem)
        self.assertIn('(:objects', pddl_problem)
        self.assertIn('l1 l2 - Location', pddl_problem)
        self.assertIn('(:init (robot_at l1) (= (battery_charge) 100))', pddl_problem)
        self.assertIn('(:goal (and (robot_at l2)))', pddl_problem)

    def test_robot_decrease_writer(self):
        problem = self.problems['robot_decrease'].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn('(:requirements :strips :typing :negative-preconditions :equality :numeric-fluents)', pddl_domain)
        self.assertIn('(:types Location)', pddl_domain)
        self.assertIn('(:predicates (robot_at ?position - Location))', pddl_domain)
        self.assertIn('(:functions (battery_charge))', pddl_domain)
        self.assertIn('(:action move', pddl_domain)
        self.assertIn(':parameters ( ?l_from - Location ?l_to - Location)', pddl_domain)
        self.assertIn(':precondition (and (<= 10 (battery_charge)) (not (= ?l_from ?l_to)) (robot_at ?l_from) (not (robot_at ?l_to)))', pddl_domain)
        self.assertIn(':effect (and (not (robot_at ?l_from)) (robot_at ?l_to) (decrease (battery_charge) 10))', pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn('(:domain robot_decrease-domain)', pddl_problem)
        self.assertIn('(:objects', pddl_problem)
        self.assertIn('l1 l2 - Location', pddl_problem)
        self.assertIn('(:init (robot_at l1) (= (battery_charge) 100))', pddl_problem)
        self.assertIn('(:goal (and (robot_at l2)))', pddl_problem)

    def test_robot_loader_writer(self):
        problem = self.problems['robot_loader'].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn('(:requirements :strips :typing :negative-preconditions :equality)', pddl_domain)
        self.assertIn('(:types Location)', pddl_domain)
        self.assertIn('(:predicates (robot_at ?position - Location) (cargo_at ?position - Location) (cargo_mounted))', pddl_domain)
        self.assertIn('(:action move', pddl_domain)
        self.assertIn(':parameters ( ?l_from - Location ?l_to - Location)', pddl_domain)
        self.assertIn(':precondition (and (not (= ?l_from ?l_to)) (robot_at ?l_from) (not (robot_at ?l_to)))', pddl_domain)
        self.assertIn(':effect (and (not (robot_at ?l_from)) (robot_at ?l_to))', pddl_domain)
        self.assertIn('(:action load', pddl_domain)
        self.assertIn(':parameters ( ?loc - Location)', pddl_domain)
        self.assertIn(':precondition (and (cargo_at ?loc) (robot_at ?loc) (not (cargo_mounted)))', pddl_domain)
        self.assertIn(':effect (and (not (cargo_at ?loc)) (cargo_mounted))', pddl_domain)
        self.assertIn('(:action unload', pddl_domain)
        self.assertIn(':parameters ( ?loc - Location)', pddl_domain)
        self.assertIn(':precondition (and (not (cargo_at ?loc)) (robot_at ?loc) (cargo_mounted))', pddl_domain)
        self.assertIn(':effect (and (cargo_at ?loc) (not (cargo_mounted)))', pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn('(:domain robot_loader-domain)', pddl_problem)
        self.assertIn('(:objects', pddl_problem)
        self.assertIn('l1 l2 - Location', pddl_problem)
        self.assertIn('(:init (robot_at l1) (cargo_at l2))', pddl_problem)
        self.assertIn('(:goal (and (cargo_at l1)))', pddl_problem)

    def test_robot_loader_adv_writer(self):
        problem = self.problems['robot_loader_adv'].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn('(:requirements :strips :typing :negative-preconditions :equality)', pddl_domain)
        self.assertIn('(:types Robot Location Container)', pddl_domain)
        self.assertIn('(:predicates (robot_at ?robot - Robot ?position - Location) (cargo_at ?cargo - Container ?position - Location) (cargo_mounted ?cargo - Container ?robot - Robot))', pddl_domain)
        self.assertIn('(:action move', pddl_domain)
        self.assertIn(':parameters ( ?l_from - Location ?l_to - Location ?r - Robot)', pddl_domain)
        self.assertIn(':precondition (and (not (= ?l_from ?l_to)) (robot_at ?r ?l_from) (not (robot_at ?r ?l_to)))', pddl_domain)
        self.assertIn(':effect (and (not (robot_at ?r ?l_from)) (robot_at ?r ?l_to))', pddl_domain)
        self.assertIn('(:action load', pddl_domain)
        self.assertIn(':parameters ( ?loc - Location ?r - Robot ?c - Container)', pddl_domain)
        self.assertIn(':precondition (and (cargo_at ?c ?loc) (robot_at ?r ?loc) (not (cargo_mounted ?c ?r)))', pddl_domain)
        self.assertIn(':effect (and (not (cargo_at ?c ?loc)) (cargo_mounted ?c ?r))', pddl_domain)
        self.assertIn('(:action unload', pddl_domain)
        self.assertIn(':parameters ( ?loc - Location ?r - Robot ?c - Container)', pddl_domain)
        self.assertIn(':precondition (and (not (cargo_at ?c ?loc)) (robot_at ?r ?loc) (cargo_mounted ?c ?r))', pddl_domain)
        self.assertIn(':effect (and (cargo_at ?c ?loc) (not (cargo_mounted ?c ?r)))', pddl_domain)

        pddl_problem = w.get_problem()
        self.assertIn('(:domain robot_loader_adv-domain)', pddl_problem)
        self.assertIn('(:objects', pddl_problem)
        self.assertIn('r1 - Robot', pddl_problem)
        self.assertIn('l1 l2 l3 - Location', pddl_problem)
        self.assertIn('c1 - Container', pddl_problem)
        self.assertIn('(:init (robot_at r1 l1) (cargo_at c1 l2))', pddl_problem)
        self.assertIn('(:goal (and (cargo_at c1 l3) (robot_at r1 l1)))', pddl_problem)

    def test_matchcellar_writer(self):
        problem = self.problems['matchcellar'].problem

        w = PDDLWriter(problem)

        pddl_domain = w.get_domain()
        self.assertIn('(define (domain MatchCellar-domain)', pddl_domain)
        self.assertIn('(:requirements :strips :typing :negative-preconditions :durative-actions)', pddl_domain)
        self.assertIn('(:types Match Fuse)', pddl_domain)
        self.assertIn('(:predicates (handfree) (light) (match_used ?match - Match) (fuse_mended ?fuse - Fuse))', pddl_domain)
        self.assertIn('(:durative-action light_match', pddl_domain)
        self.assertIn(':parameters ( ?m - Match)', pddl_domain)
        self.assertIn(':duration (= ?duration 6)', pddl_domain)
        self.assertIn(':condition (and (at start (not (match_used ?m))))', pddl_domain)
        self.assertIn(':effect (and (at start (match_used ?m)) (at start (light)) (at end (not (light)))))', pddl_domain)
        self.assertIn('(:durative-action mend_fuse', pddl_domain)
        self.assertIn(':parameters ( ?f - Fuse)', pddl_domain)
        self.assertIn(':duration (= ?duration 5)', pddl_domain)
        self.assertIn(':condition (and (at start (handfree))(at start (light))(over all (light))(at end (light)))', pddl_domain)
        self.assertIn(':effect (and (at start (not (handfree))) (at end (fuse_mended ?f)) (at end (handfree))))', pddl_domain)

    def test_depot_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(PDDL_DOMAINS_PATH, 'depot', 'domain.pddl')
        problem_filename = os.path.join(PDDL_DOMAINS_PATH, 'depot', 'problem.pddl')
        problem = reader.parse_problem(domain_filename, problem_filename)

        self.assertTrue(problem is not None)
        self.assertEqual(len(problem.fluents), 15)
        self.assertEqual(len(problem.actions), 5)
        self.assertEqual(len(list(problem.objects(problem.user_type('object')))), 13)

    def test_counters_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(PDDL_DOMAINS_PATH, 'counters', 'domain.pddl')
        problem_filename = os.path.join(PDDL_DOMAINS_PATH, 'counters', 'problem.pddl')
        problem = reader.parse_problem(domain_filename, problem_filename)

        self.assertTrue(problem is not None)
        self.assertEqual(len(problem.fluents), 2)
        self.assertEqual(len(problem.actions), 2)
        self.assertEqual(len(list(problem.objects(problem.user_type('counter')))), 4)

    def test_sailing_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(PDDL_DOMAINS_PATH, 'sailing', 'domain.pddl')
        problem_filename = os.path.join(PDDL_DOMAINS_PATH, 'sailing', 'problem.pddl')
        problem = reader.parse_problem(domain_filename, problem_filename)

        self.assertTrue(problem is not None)
        self.assertEqual(len(problem.fluents), 4)
        self.assertEqual(len(problem.actions), 8)
        self.assertEqual(len(list(problem.objects(problem.user_type('boat')))), 2)
        self.assertEqual(len(list(problem.objects(problem.user_type('person')))), 2)

    def test_matchcellar_reader(self):
        reader = PDDLReader()

        domain_filename = os.path.join(PDDL_DOMAINS_PATH, 'matchcellar', 'domain.pddl')
        problem_filename = os.path.join(PDDL_DOMAINS_PATH, 'matchcellar', 'problem.pddl')
        problem = reader.parse_problem(domain_filename, problem_filename)

        self.assertTrue(problem is not None)
        self.assertEqual(len(problem.fluents), 4)
        self.assertEqual(len(problem.actions), 2)
        self.assertEqual(len(list(problem.objects(problem.user_type('match')))), 3)
        self.assertEqual(len(list(problem.objects(problem.user_type('fuse')))), 3)

    def test_examples_io(self):
        for example in self.problems.values():
            problem = example.problem
            kind = problem.kind
            if kind.has_intermediate_conditions_and_effects() or \
                kind.has_object_fluents():
                continue
            with tempfile.TemporaryDirectory() as tempdir:
                domain_filename = os.path.join(tempdir, 'domain.pddl')
                problem_filename = os.path.join(tempdir, 'problem.pddl')

                w = PDDLWriter(problem)
                w.write_domain(domain_filename)
                w.write_problem(problem_filename)

                reader = PDDLReader()
                parsed_problem = reader.parse_problem(domain_filename, problem_filename)

                if problem.has_type('object') and problem.kind.has_hierarchical_typing():
                    object_rename: str = 'object'
                    while problem.has_type(object_rename):
                        object_rename = f'{object_rename}_'
                    self.assertEqual(len(problem.fluents), len(parsed_problem.fluents))
                    self.assertTrue(_have_same_user_types_considering_object_renaming(problem, parsed_problem, object_rename))
                    self.assertEqual(len(problem.actions), len(parsed_problem.actions))
                    for a in problem.actions:
                        parsed_a = parsed_problem.action(a.name)
                        self.assertEqual(a.name, parsed_a.name)
                        for param, parsed_param in zip(a.parameters, parsed_a.parameters):
                            self.assertEqual(param.name, parsed_param.name)
                            self.assertTrue(_is_same_user_type_considering_object_renaming(param.type, parsed_param.type, object_rename))
                        if isinstance(a, unified_planning.model.InstantaneousAction):
                            self.assertEqual(len(a.effects), len(parsed_a.effects))
                        elif isinstance(a, unified_planning.model.DurativeAction):
                            self.assertEqual(a.duration, parsed_a.duration)
                            for t, e in a.effects.items():
                                self.assertEqual(len(e), len(parsed_a.effects[t]))
                else:
                    self.assertEqual(len(problem.fluents), len(parsed_problem.fluents))
                    self.assertEqual(set(problem.user_types), set(parsed_problem.user_types))
                    self.assertEqual(len(problem.actions), len(parsed_problem.actions))
                    for a in problem.actions:
                        parsed_a = parsed_problem.action(a.name)
                        self.assertEqual(a.name, parsed_a.name)
                        self.assertEqual(a.parameters, parsed_a.parameters)
                        if isinstance(a, unified_planning.model.InstantaneousAction):
                            self.assertEqual(len(a.effects), len(parsed_a.effects))
                        elif isinstance(a, unified_planning.model.DurativeAction):
                            self.assertEqual(a.duration, parsed_a.duration)
                            for t, e in a.effects.items():
                                self.assertEqual(len(e), len(parsed_a.effects[t]))

                    self.assertEqual(set(problem.all_objects), set(parsed_problem.all_objects))
                    self.assertEqual(len(problem.initial_values), len(parsed_problem.initial_values))

    def test_rationals(self):
        problem = self.problems['robot_decrease'].problem.clone()

        # Check perfect conversion
        battery = problem.fluent('battery_charge')
        problem.set_initial_value(battery, Fraction(5, 2))
        w = PDDLWriter(problem)
        pddl_txt = w.get_problem()
        self.assertNotIn('5/2', pddl_txt)
        self.assertIn('2.5', pddl_txt)

        # Check imperfect conversion
        with pytest.warns(UserWarning, match="cannot exactly represent") as warns:
            battery = problem.fluent('battery_charge')
            problem.set_initial_value(battery, Fraction(10, 3))
            w = PDDLWriter(problem)
            pddl_txt = w.get_problem()
            self.assertNotIn('10/3', pddl_txt)
            self.assertIn('3.333333333', pddl_txt)

def _is_same_user_type_considering_object_renaming(original_type: unified_planning.model.Type,
                                                    tested_type: unified_planning.model.Type,
                                                    object_rename: str) -> bool:
    if cast(_UserType, original_type).father is None: # case where original_type has no father
        if cast(_UserType, tested_type).father is not None:
            return False # original_type has a father, tested_type does not.
        if cast(_UserType, original_type).name != 'object': # fathers are both None, now we have to check the name
            return cast(_UserType, original_type).name == cast(_UserType, tested_type).name # original_type name is not object, so it should not be renamed
        else:
            return object_rename == cast(_UserType, tested_type).name # original_type name is object, so we expect it to be renamed
    else: # case where original_type has a father
        if cast(_UserType, tested_type).father is None:
            return False # and tested_type has no father
        if cast(_UserType, original_type).name != 'object': # original_type name is not object, so it should not be renamed, and both father's types are the same type
            return cast(_UserType, original_type).name == cast(_UserType, tested_type).name and _is_same_user_type_considering_object_renaming(original_type.father, tested_type.father, object_rename) # type: ignore
        else: # original_type name is object, so we expect it to be renamed, and both father's types are the same type
            return object_rename == tested_type.name and _is_same_user_type_considering_object_renaming(original_type.father, tested_type.father, object_rename) # type: ignore

def _have_same_user_types_considering_object_renaming(original_problem: unified_planning.model.Problem, tested_problem: unified_planning.model.Problem, object_rename: str) -> bool:
    for original_type in original_problem.user_types:
        if cast(_UserType, original_type).name != 'object':
            if not _is_same_user_type_considering_object_renaming(original_type, tested_problem.user_type(cast(_UserType, original_type).name), object_rename):
                return False
        else:
            if not _is_same_user_type_considering_object_renaming(original_type, tested_problem.user_type(object_rename), object_rename):
                return False
    return True



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines all the types."""

import unified_planning
from fractions import Fraction
from typing import Iterator, Optional, Dict, Tuple, List
from unified_planning.exceptions import UPProblemDefinitionError, UPTypeError


class Type:
    """Basic class for representing a type."""

    def is_bool_type(self) -> bool:
        """Returns true iff is boolean type."""
        return False

    def is_user_type(self) -> bool:
        """Returns true iff is a user type."""
        return False

    def is_real_type(self) -> bool:
        """Returns true iff is real type."""
        return False

    def is_int_type(self) -> bool:
        """Returns true iff is integer type."""
        return False


class _BoolType(Type):
    """Represents the boolean type."""

    def __repr__(self) -> str:
        return 'bool'

    def is_bool_type(self) -> bool:
        """Returns true iff is boolean type."""
        return True


class _UserType(Type):
    """Represents the user type."""
    def __init__(self, name: str, father: Optional[Type] = None):
        Type.__init__(self)
        self._name = name
        if father is not None and (not father.is_user_type()):
            raise UPTypeError('father field of a UserType must be a UserType.')
        self._father = father

    def __repr__(self) -> str:
        return self._name if self._father is None else f'{self._name} - {self._father.name}' # type: ignore

    @property
    def name(self) -> str:
        """Returns the type name."""
        return self._name

    @property
    def father(self) -> Optional[Type]:
        """Returns the type s father."""
        return self._father

    def is_user_type(self) -> bool:
        """Returns true iff is a user type."""
        return True


class _IntType(Type):
    def __init__(self, lower_bound: int = None, upper_bound: int = None):
        Type.__init__(self)
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def __repr__(self) -> str:
        b = []
        if (not self.lower_bound is None) or (not self.upper_bound is None):
            b.append('[')
            b.append('-inf' if self.lower_bound is None else str(self.lower_bound))
            b.append(', ')
            b.append('inf' if self.upper_bound is None else str(self.upper_bound))
            b.append(']')
        return 'integer' + ''.join(b)

    @property
    def lower_bound(self) -> Optional[int]:
        return self._lower_bound

    @property
    def upper_bound(self) -> Optional[int]:
        return self._upper_bound

    def is_int_type(self) -> bool:
        return True


class _RealType(Type):
    def __init__(self, lower_bound: Fraction = None, upper_bound: Fraction = None):
        Type.__init__(self)
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def __repr__(self) -> str:
        b = []
        if (not self.lower_bound is None) or (not self.upper_bound is None):
            b.append('[')
            b.append('-inf' if self.lower_bound is None else str(self.lower_bound))
            b.append(', ')
            b.append('inf' if self.upper_bound is None else str(self.upper_bound))
            b.append(']')
        return 'real' + ''.join(b)

    @property
    def lower_bound(self) -> Optional[Fraction]:
        return self._lower_bound

    @property
    def upper_bound(self) -> Optional[Fraction]:
        return self._upper_bound

    def is_real_type(self) -> bool:
        return True


BOOL = _BoolType()

class TypeManager:
    def __init__(self):
        self._bool = BOOL
        self._ints: Dict[Tuple[Optional[int], Optional[int]], Type] = {}
        self._reals: Dict[Tuple[Optional[Fraction], Optional[Fraction]], Type] = {}
        self._user_types: Dict[Tuple[str, Optional[Type]], Type] = {}

    def BoolType(self) -> Type:
        return self._bool

    def IntType(self, lower_bound: int = None, upper_bound: int = None) -> Type:
        k = (lower_bound, upper_bound)
        if k in self._ints:
            return self._ints[k]
        else:
            it = _IntType(lower_bound, upper_bound)
            self._ints[k] = it
            return it

    def RealType(self, lower_bound: Fraction = None, upper_bound: Fraction = None) -> Type:
        k = (lower_bound, upper_bound)
        if k in self._reals:
            return self._reals[k]
        else:
            rt = _RealType(lower_bound, upper_bound)
            self._reals[k] = rt
            return rt

    def UserType(self, name: str, father: Optional[Type] = None) -> Type:
        if (name, father) in self._user_types:
            return self._user_types[(name, father)]
        else:
            if father is not None:
                if any(ancestor.name == name for ancestor in self.user_type_ancestors(father)): # type: ignore
                    raise UPTypeError('The name: {name} is already used in the UserType: {ancestor}. An UserType and one of his ancestors can not share the name.')
            ut = _UserType(name, father)
            self._user_types[(name, father)] = ut
            return ut

    def user_type_ancestors(self, user_type: Type) -> Iterator[Type]:
        '''Returns all the ancestors of the given UserType, including itself.'''
        if not user_type.is_user_type():
            raise UPTypeError('The function user_type_ancestors can be called only on UserTypes.')
        yield user_type
        father: Optional[Type] = user_type.father # type: ignore
        while father is not None:
            yield father
            father = father.father # type: ignore


def domain_size(problem: 'unified_planning.model.problem.Problem', typename: 'unified_planning.model.types.Type') -> int:
    '''Returns the domain size of the given type.'''
    if typename.is_bool_type():
        return 2
    elif typename.is_user_type():
        return len(list(problem.objects_hierarchy(typename)))
    elif typename.is_int_type():
        lb = typename.lower_bound # type: ignore
        ub = typename.upper_bound # type: ignore
        if lb is None or ub is None:
            raise UPProblemDefinitionError('Parameter not groundable!')
        return ub - lb
    else:
        raise UPProblemDefinitionError('Parameter not groundable!')

def domain_item(problem: 'unified_planning.model.problem.Problem', typename: 'unified_planning.model.types.Type', idx: int) -> 'unified_planning.model.fnode.FNode':
    '''Returns the ith domain item of the given type.'''
    if typename.is_bool_type():
        return problem._env.expression_manager.Bool(idx == 0)
    elif typename.is_user_type():
        return problem._env.expression_manager.ObjectExp(list(problem.objects_hierarchy(typename))[idx])
    elif typename.is_int_type():
        lb = typename.lower_bound # type: ignore
        ub = typename.upper_bound # type: ignore
        if lb is None or ub is None:
            raise UPProblemDefinitionError('Parameter not groundable!')
        return problem._env.expression_manager.Int(lb + idx)
    else:
        raise UPProblemDefinitionError('Parameter not groundable!')



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
import unified_planning.model
import pyparsing # type: ignore
import typing
from mimetypes import types_map
from unified_planning.environment import Environment, get_env
from collections import OrderedDict
from fractions import Fraction
from typing import Dict, Union, Callable, List, cast
from pyparsing import Word, alphanums, alphas, ZeroOrMore, OneOrMore, Keyword
from pyparsing import Optional, Suppress, nestedExpr, Group, restOfLine
if pyparsing.__version__ < '3.0.0':
    from pyparsing import oneOf as one_of
    from pyparsing import ParseResults
else:
    from pyparsing.results import ParseResults # type: ignore
    from pyparsing import one_of


class PDDLGrammar:
    def __init__(self):
        name = Word(alphas, alphanums+'_'+'-')
        variable = Suppress('?') + name

        require_def = Suppress('(') + ':requirements' + \
            OneOrMore(one_of(':strips :typing :negative-preconditions :disjunctive-preconditions :equality :existential-preconditions :universal-preconditions :quantified-preconditions :conditional-effects :fluents :numeric-fluents :adl :durative-actions :duration-inequalities :timed-initial-literals :action-costs')) \
            + Suppress(')')

        types_def = Suppress('(') + ':types' + \
            OneOrMore(Group(Group(OneOrMore(name)) + \
                            Optional(Suppress('-') + name))).setResultsName('types') \
                            + Suppress(')')

        constants_def = Suppress('(') + ':constants' + \
            OneOrMore(Group(Group(OneOrMore(name)) + \
                            Optional(Suppress('-') + name))).setResultsName('constants') \
                            + Suppress(')')

        predicate = Suppress('(') + \
            Group(name + Group(ZeroOrMore(Group(Group(OneOrMore(variable)) + \
                                                Optional(Suppress('-') + name))))) \
                                                + Suppress(')')

        predicates_def = Suppress('(') + ':predicates' + \
            Group(OneOrMore(predicate)).setResultsName('predicates') \
            + Suppress(')')

        functions_def = Suppress('(') + ':functions' + \
            Group(OneOrMore(predicate)).setResultsName('functions') \
            + Suppress(')')

        parameters = ZeroOrMore(Group(Group(OneOrMore(variable)) \
                                     + Optional(Suppress('-') + name))).setResultsName('params')
        action_def = Group(Suppress('(') + ':action' + name.setResultsName('name') \
                           + ':parameters' + Suppress('(') + parameters + Suppress(')') \
                           + Optional(':precondition' + nestedExpr().setResultsName('pre')) \
                           + Optional(':effect' + nestedExpr().setResultsName('eff')) \
                           + Suppress(')'))

        dur_action_def = Group(Suppress('(') + ':durative-action' + name.setResultsName('name') \
                               + ':parameters' + Suppress('(') + parameters + Suppress(')') \
                               + ':duration' + nestedExpr().setResultsName('duration') \
                               + ':condition' + nestedExpr().setResultsName('cond') \
                               + ':effect' + nestedExpr().setResultsName('eff') \
                               + Suppress(')'))

        domain = Suppress('(') + 'define' \
            + Suppress('(') + 'domain' + name.setResultsName('name') + Suppress(')') \
            + Optional(require_def) + Optional(types_def) + Optional(constants_def) \
            + Optional(predicates_def) + Optional(functions_def) \
            + Group(ZeroOrMore(action_def | dur_action_def)).setResultsName('actions') + Suppress(')')

        objects = OneOrMore(Group(Group(OneOrMore(name)) \
                                  + Optional(Suppress('-') + name))).setResultsName('objects')

        metric = (Keyword('minimize') | Keyword('maximize')).setResultsName('optimization') \
            + (name | nestedExpr()).setResultsName('metric')

        problem = Suppress('(') + 'define' \
            + Suppress('(') + 'problem' + name.setResultsName('name') + Suppress(')') \
            + Suppress('(') + ':domain' + name + Suppress(')') + Optional(require_def) \
            + Optional(Suppress('(') + ':objects' + objects + Suppress(')')) \
            + Suppress('(') + ':init' + ZeroOrMore(nestedExpr()).setResultsName('init') + Suppress(')') \
            + Suppress('(') + ':goal' + nestedExpr().setResultsName('goal') + Suppress(')') \
            + Optional(Suppress('(') + ':metric' + metric + Suppress(')')) \
            + Suppress(')')

        domain.ignore(';' + restOfLine)
        problem.ignore(';' + restOfLine)

        self._domain = domain
        self._problem = problem
        self._parameters = parameters

    @property
    def domain(self):
        return self._domain

    @property
    def problem(self):
        return self._problem

    @property
    def parameters(self):
        return self._parameters


class PDDLReader:
    """
    Parse a PDDL problem and generate a unified_planning problem.
    """
    def __init__(self, env: Environment = None):
        self._env = get_env(env)
        self._em = self._env.expression_manager
        self._tm = self._env.type_manager
        self._operators: Dict[str, Callable] = {
            'and' : self._em.And,
            'or' : self._em.Or,
            'not' : self._em.Not,
            'imply' : self._em.Implies,
            '>=' : self._em.GE,
            '<=' : self._em.LE,
            '>' : self._em.GT,
            '<' : self._em.LT,
            '=' : self._em.Equals,
            '+' : self._em.Plus,
            '-' : self._em.Minus,
            '/' : self._em.Div,
            '*' : self._em.Times
        }
        grammar = PDDLGrammar()
        self._pp_domain = grammar.domain
        self._pp_problem = grammar.problem
        self._pp_parameters = grammar.parameters
        self._fve = up.walkers.FreeVarsExtractor()
        self._totalcost: Optional[up.model.FNode] = None

    def _parse_exp(self, problem: up.model.Problem, act: typing.Optional[up.model.Action],
                   types_map: Dict[str, up.model.Type], var: Dict[str, up.model.Variable],
                   exp: Union[ParseResults, str]) -> up.model.FNode:
        stack = [(var, exp, False)]
        solved: List[up.model.FNode] = []
        while len(stack) > 0:
            var, exp, status = stack.pop()
            if status:
                if exp[0] == '-' and len(exp) == 2: # unary minus
                    solved.append(self._em.Times(-1, solved.pop()))
                elif exp[0] in self._operators: # n-ary operators
                    op: Callable = self._operators[exp[0]]
                    solved.append(op(*[solved.pop() for _ in exp[1:]]))
                elif exp[0] in ['exists', 'forall']: # quantifier operators
                    q_op: Callable = self._em.Exists if exp[0] == 'exists' else self._em.Forall
                    solved.append(q_op(solved.pop(), *var.values()))
                elif problem.has_fluent(exp[0]): # fluent reference
                    f = problem.fluent(exp[0])
                    args = [solved.pop() for _ in exp[1:]]
                    solved.append(self._em.FluentExp(f, tuple(args)))
                else:
                    raise up.exceptions.UPUnreachableCodeError
            else:
                if isinstance(exp, ParseResults):
                    if len(exp) == 0: # empty precodition
                        solved.append(self._em.TRUE())
                    elif exp[0] == '-' and len(exp) == 2: # unary minus
                        stack.append((var, exp, True))
                        stack.append((var, exp[1], False))
                    elif exp[0] in self._operators: # n-ary operators
                        stack.append((var, exp, True))
                        for e in exp[1:]:
                            stack.append((var, e, False))
                    elif exp[0] in ['exists', 'forall']: # quantifier operators
                        vars_string = ' '.join(exp[1])
                        vars_res = self._pp_parameters.parseString(vars_string)
                        vars = {}
                        for g in vars_res['params']:
                            t = types_map[g[1] if len(g) > 1 else 'object']
                            for o in g[0]:
                                vars[o] = up.model.Variable(o, t)
                        stack.append((vars, exp, True))
                        stack.append((vars, exp[2], False))
                    elif problem.has_fluent(exp[0]): # fluent reference
                        stack.append((var, exp, True))
                        for e in exp[1:]:
                            stack.append((var, e, False))
                    elif len(exp) == 1: # expand an element inside brackets
                        stack.append((var, exp[0], False))
                    else:
                        raise SyntaxError(f'Not able to handle: {exp}')
                elif isinstance(exp, str):
                    if exp[0] == '?' and exp[1:] in var: # variable in a quantifier expression
                        solved.append(self._em.VariableExp(var[exp[1:]]))
                    elif exp[0] == '?': # action parameter
                        assert act is not None
                        solved.append(self._em.ParameterExp(act.parameter(exp[1:])))
                    elif problem.has_fluent(exp): # fluent
                        solved.append(self._em.FluentExp(problem.fluent(exp)))
                    elif problem.has_object(exp): # object
                        solved.append(self._em.ObjectExp(problem.object(exp)))
                    else: # number
                        n = Fraction(exp)
                        if n.denominator == 1:
                            solved.append(self._em.Int(n.numerator))
                        else:
                            solved.append(self._em.Real(n))
                else:
                    raise SyntaxError(f'Not able to handle: {exp}')
        assert len(solved) == 1 #sanity check
        return solved.pop()

    def _add_effect(self, problem: up.model.Problem,
                    act: Union[up.model.InstantaneousAction, up.model.DurativeAction],
                    types_map: Dict[str, up.model.Type],
                    exp: Union[ParseResults, str],
                    cond: Union[up.model.FNode, bool] = True,
                    timing: typing.Optional[up.model.Timing] = None):
        to_add = [(exp, cond)]
        while to_add:
            exp, cond = to_add.pop(0)
            op = exp[0]
            if op == 'and':
                exp = exp[1:]
                for e in exp:
                    to_add.append((e, cond))
            elif op == 'when':
                cond = self._parse_exp(problem, act, types_map, {}, exp[1])
                to_add.append((exp[2], cond))
            elif op == 'not':
                exp = exp[1]
                eff = (self._parse_exp(problem, act, types_map, {}, exp), self._em.FALSE(), cond)
                act.add_effect(*eff if timing is None else (timing, *eff)) # type: ignore
            elif op == 'assign':
                eff = (self._parse_exp(problem, act, types_map, {}, exp[1]),
                       self._parse_exp(problem, act, types_map, {}, exp[2]), cond)
                act.add_effect(*eff if timing is None else (timing, *eff)) # type: ignore
            elif op == 'increase':
                eff = (self._parse_exp(problem, act, types_map, {}, exp[1]),
                       self._parse_exp(problem, act, types_map, {}, exp[2]), cond)
                act.add_increase_effect(*eff if timing is None else (timing, *eff)) # type: ignore
            elif op == 'decrease':
                eff = (self._parse_exp(problem, act, types_map, {}, exp[1]),
                       self._parse_exp(problem, act, types_map, {}, exp[2]), cond)
                act.add_decrease_effect(*eff if timing is None else (timing, *eff)) # type: ignore
            else:
                eff = (self._parse_exp(problem, act, types_map, {}, exp), self._em.TRUE(), cond)
                act.add_effect(*eff if timing is None else (timing, *eff)) # type: ignore

    def _add_condition(self, problem: up.model.Problem, act: up.model.DurativeAction,
                       exp: Union[ParseResults, str],
                       types_map: Dict[str, up.model.Type],
                       vars: typing.Optional[Dict[str, up.model.Variable]] = None):
        to_add = [(exp, vars)]
        while to_add:
            exp, vars = to_add.pop(0)
            op = exp[0]
            if op == 'and':
                for e in exp[1:]:
                    to_add.append((e, vars))
            elif op == 'forall':
                vars_string = ' '.join(exp[1])
                vars_res = self._pp_parameters.parseString(vars_string)
                if vars is None:
                    vars = {}
                for g in vars_res['params']:
                    t = types_map[g[1] if len(g) > 1 else 'object']
                    for o in g[0]:
                        vars[o] = up.model.Variable(o, t)
                to_add.append((exp[2], vars))
            elif len(exp) == 3 and op == 'at' and exp[1] == 'start':
                cond = self._parse_exp(problem, act, types_map, {} if vars is None else vars, exp[2])
                if vars is not None:
                    cond = self._em.Forall(cond, *vars.values())
                act.add_condition(up.model.StartTiming(), cond)
            elif len(exp) == 3 and op == 'at' and exp[1] == 'end':
                cond = self._parse_exp(problem, act, types_map, {} if vars is None else vars, exp[2])
                if vars is not None:
                    cond = self._em.Forall(cond, *vars.values())
                act.add_condition(up.model.EndTiming(), cond)
            elif len(exp) == 3 and op == 'over' and exp[1] == 'all':
                t_all = up.model.OpenTimeInterval(up.model.StartTiming(), up.model.EndTiming())
                cond = self._parse_exp(problem, act, types_map, {} if vars is None else vars, exp[2])
                if vars is not None:
                    cond = self._em.Forall(cond, *vars.values())
                act.add_condition(t_all, cond)
            else:
                raise SyntaxError(f'Not able to handle: {exp}')

    def _add_timed_effects(self, problem: up.model.Problem,
                           act: up.model.DurativeAction,
                           types_map: Dict[str, up.model.Type],
                           eff: ParseResults):
        to_add = [eff]
        while to_add:
            eff = to_add.pop(0)
            op = eff[0]
            if op == 'and':
                for e in eff[1:]:
                    to_add.append(e)
            elif len(eff) == 3 and op == 'at' and eff[1] == 'start':
                self._add_effect(problem, act, types_map, eff[2], timing=up.model.StartTiming())
            elif len(eff) == 3 and op == 'at' and eff[1] == 'end':
                self._add_effect(problem, act, types_map, eff[2], timing=up.model.EndTiming())
            else:
                raise SyntaxError(f'Not able to handle: {eff}')

    def _check_if_object_type_is_needed(self, domain_res) -> bool:
        for p in domain_res.get('predicates', []):
            for g in p[1]:
                if len(g) <= 1 or g[1] == 'object':
                    return True
        for p in domain_res.get('functions', []):
            for g in p[1]:
                if len(g) <= 1 or g[1] == 'object':
                    return True
        for g in domain_res.get('constants', []):
            if len(g) <= 1 or g[1] == 'object':
                return True
        for a in domain_res.get('actions', []):
            for g in a.get('params', []):
                if len(g) <= 1 or g[1] == 'object':
                    return True
        return False

    def _durative_action_has_cost(self, dur_act: up.model.DurativeAction):
        if self._totalcost in self._fve.get(dur_act.duration.lower) or \
           self._totalcost in self._fve.get(dur_act.duration.upper):
            return False
        for _, cl in dur_act.conditions.items():
            for c in cl:
                if self._totalcost in self._fve.get(c):
                    return False
        for _, el in dur_act.effects.items():
            for e in el:
                if self._totalcost in self._fve.get(e.fluent) or self._totalcost in self._fve.get(e.value) \
                   or self._totalcost in self._fve.get(e.condition):
                    return False
        return True

    def _instantaneous_action_has_cost(self, act: up.model.InstantaneousAction):
        for c in act.preconditions:
            if self._totalcost in self._fve.get(c):
                return False
        for e in act.effects:
            if self._totalcost in self._fve.get(e.value) or self._totalcost in self._fve.get(e.condition):
                return False
            if e.fluent == self._totalcost:
                if not e.is_increase() or \
                   not e.condition.is_true() or \
                   not (e.value.is_int_constant() or \
                        e.value.is_real_constant()):
                    return False
        return True

    def _problem_has_actions_cost(self, problem: up.model.Problem):
        if self._totalcost is None or not problem.initial_value(self._totalcost).constant_value() == 0:
            return False
        for _, el in problem.timed_effects.items():
            for e in el:
                if self._totalcost in self._fve.get(e.fluent) or self._totalcost in self._fve.get(e.value) \
                   or self._totalcost in self._fve.get(e.condition):
                    return False
        for c in problem.goals:
            if self._totalcost in self._fve.get(c):
                return False
        return True

    def parse_problem(self, domain_filename: str,
                      problem_filename: typing.Optional[str] = None) -> 'up.model.Problem':
        domain_res = self._pp_domain.parseFile(domain_filename)

        problem = up.model.Problem(domain_res['name'], self._env,
                                   initial_defaults={self._tm.BoolType(): self._em.FALSE()})

        types_map: Dict[str, 'up.model.Type'] = {}
        object_type_needed: bool = self._check_if_object_type_is_needed(domain_res)
        for types_list in domain_res.get('types', []):
            # types_list is a List of 1 or 2 elements, where the first one
            # is a List of types, and the second one can be their father,
            # if they have one.
            father: Optional['up.model.Type'] = None
            if len(types_list) == 2: # the types have a father
                if types_list[1] != 'object': #the father is not object
                    father = types_map[types_list[1]]
                elif object_type_needed: # the father is object, and object is needed
                    object_type = types_map.get('object', None)
                    if object_type is None: # the type object is not defined
                        father = self._env.type_manager.UserType('object', None)
                        types_map['object'] = father
                    else:
                        father = object_type
            else:
                assert len(types_list) == 1, "Malformed list of types, I was expecting either 1 or 2 elements" # sanity check
            for type_name in types_list[0]:
                types_map[type_name] = self._env.type_manager.UserType(type_name, father)
        if object_type_needed and 'object' not in types_map: # The object type is needed, but has not been defined
            types_map['object'] = self._env.type_manager.UserType('object', None)  # We manually define it.

        has_actions_cost = False

        for p in domain_res.get('predicates', []):
            n = p[0]
            params = OrderedDict()
            for g in p[1]:
                param_type = types_map[g[1] if len(g) > 1 else 'object']
                for param_name in g[0]:
                    params[param_name] = param_type
            f = up.model.Fluent(n, self._tm.BoolType(), params, self._env)
            problem.add_fluent(f)

        for p in domain_res.get('functions', []):
            n = p[0]
            params = OrderedDict()
            for g in p[1]:
                param_type = types_map[g[1] if len(g) > 1 else 'object']
                for param_name in g[0]:
                    params[param_name] = param_type
            f = up.model.Fluent(n, self._tm.RealType(), params, self._env)
            if n == 'total-cost':
                has_actions_cost = True
                self._totalcost = cast(up.model.FNode, self._em.FluentExp(f))
            problem.add_fluent(f)

        for g in domain_res.get('constants', []):
            t = types_map[g[1] if len(g) > 1 else 'object']
            for o in g[0]:
                problem.add_object(up.model.Object(o, t))

        for a in domain_res.get('actions', []):
            n = a['name']
            a_params = OrderedDict()
            for g in a.get('params', []):
                t = types_map[g[1] if len(g) > 1 else 'object']
                for p in g[0]:
                    a_params[p] = t
            if 'duration' in a:
                dur_act = up.model.DurativeAction(n, a_params, self._env)
                dur = a['duration'][0]
                if dur[0] == '=':
                    dur_act.set_fixed_duration(self._parse_exp(problem, dur_act, types_map,
                                                               {}, dur[2]))
                elif dur[0] == 'and':
                    upper = None
                    lower = None
                    for j in range(1, len(dur)):
                        v = Fraction(dur[j][2])
                        if dur[j][0] == '>=':
                            if lower is None or v > lower:
                                lower = v
                        elif dur[j][0] == '<=':
                            if upper is None or v < upper:
                                upper = v
                        else:
                            raise SyntaxError(f'Not able to handle duration constraint of action {n}')
                    if lower is None or upper is None:
                        raise SyntaxError(f'Not able to handle duration constraint of action {n}')
                    d = up.model.ClosedDurationInterval(self._em.Real(lower),
                                                        self._em.Real(upper))
                    dur_act.set_duration_constraint(d)
                else:
                    raise SyntaxError(f'Not able to handle duration constraint of action {n}')
                cond = a['cond'][0]
                self._add_condition(problem, dur_act, cond, types_map)
                eff = a['eff'][0]
                self._add_timed_effects(problem, dur_act, types_map, eff)
                problem.add_action(dur_act)
                has_actions_cost = has_actions_cost and self._durative_action_has_cost(dur_act)
            else:
                act = up.model.InstantaneousAction(n, a_params, self._env)
                if 'pre' in a:
                    act.add_precondition(self._parse_exp(problem, act, types_map, {}, a['pre'][0]))
                if 'eff' in a:
                    self._add_effect(problem, act, types_map, a['eff'][0])
                problem.add_action(act)
                has_actions_cost = has_actions_cost and self._instantaneous_action_has_cost(act)

        if problem_filename is not None:
            problem_res = self._pp_problem.parseFile(problem_filename)

            problem.name = problem_res['name']

            for g in problem_res.get('objects', []):
                t = types_map[g[1] if len(g) > 1 else 'object']
                for o in g[0]:
                    problem.add_object(up.model.Object(o, t))

            for i in problem_res.get('init', []):
                if i[0] == '=':
                    problem.set_initial_value(self._parse_exp(problem, None, types_map, {}, i[1]),
                                              self._parse_exp(problem, None, types_map, {}, i[2]))
                elif len(i) == 3 and i[0] == 'at' and i[1].replace('.','',1).isdigit():
                    ti = up.model.StartTiming(Fraction(i[1]))
                    va = self._parse_exp(problem, None, types_map, {}, i[2])
                    if va.is_fluent_exp():
                        problem.add_timed_effect(ti, va, self._em.TRUE())
                    elif va.is_not():
                        problem.add_timed_effect(ti, va.arg(0), self._em.FALSE())
                    elif va.is_equals():
                        problem.add_timed_effect(ti, va.arg(0), va.arg(1))
                    else:
                        raise SyntaxError(f'Not able to handle this TIL {i}')
                else:
                    problem.set_initial_value(self._parse_exp(problem, None, types_map, {}, i), self._em.TRUE())

            problem.add_goal(self._parse_exp(problem, None, types_map, {}, problem_res['goal'][0]))

            has_actions_cost = has_actions_cost and self._problem_has_actions_cost(problem)

            optimization = problem_res.get('optimization', None)
            metric = problem_res.get('metric', None)

            if metric is not None:
                if optimization == 'minimize' and len(metric) == 1 and metric[0] == 'total-time':
                    problem.add_quality_metric(up.model.metrics.MinimizeMakespan())
                else:
                    metric_exp = self._parse_exp(problem, None, types_map, {}, metric)
                    if has_actions_cost and optimization == 'minimize' and metric_exp == self._totalcost:
                        costs = {}
                        problem._fluents.remove(self._totalcost.fluent())
                        problem._initial_value.pop(self._totalcost)
                        use_plan_length = all(False for _ in problem.durative_actions)
                        for a in problem.instantaneous_actions:
                            cost = None
                            for e in a.effects:
                                if e.fluent == self._totalcost:
                                    cost = e
                                    break
                            if cost is not None:
                                costs[a] = cost.value
                                a._effects.remove(cost)
                                if cost.value != 1:
                                    use_plan_length = False
                            else:
                                use_plan_length = False
                        if use_plan_length:
                            problem.add_quality_metric(up.model.metrics.MinimizeSequentialPlanLength())
                        else:
                            problem.add_quality_metric(up.model.metrics.MinimizeActionCosts(costs, self._em.Int(0)))
                    else:
                        if optimization == 'minimize':
                            problem.add_quality_metric(up.model.metrics.MinimizeExpressionOnFinalState(metric_exp))
                        elif optimization == 'maximize':
                            problem.add_quality_metric(up.model.metrics.MaximizeExpressionOnFinalState(metric_exp))

        return problem



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the negative preconditions remover class."""



from unified_planning.transformers.transformer import Transformer
from unified_planning.model import Fluent, Problem, InstantaneousAction, DurativeAction, FNode, Action, Effect, Timing
from unified_planning.walkers.identitydag import IdentityDagWalker
from unified_planning.exceptions import UPExpressionDefinitionError, UPProblemDefinitionError
from typing import List, Dict, Union

class NegativeFluentRemover(IdentityDagWalker):
    def __init__(self, negative_conditions_remover, env):
        self._env = env
        IdentityDagWalker.__init__(self, self._env)
        self._fluent_mapping: Dict[Fluent, Fluent] = {}
        self._negative_conditions_remover = negative_conditions_remover

    def remove_negative_fluents(self, expression:FNode) -> FNode:
        return self.walk(expression)

    def walk_not(self, expression: FNode, args: List[FNode], **kwargs) -> FNode:
        assert len(args) == 1
        if not args[0].is_fluent_exp():
            raise UPExpressionDefinitionError(f"Expression: {expression} is not in NNF.")
        neg_fluent = self._fluent_mapping.get(args[0].fluent(), None)
        if neg_fluent is not None:
            return self._env.expression_manager.FluentExp(neg_fluent, tuple(args[0].args))
        f = args[0].fluent()
        nf = Fluent(self._negative_conditions_remover.get_fresh_name(f.name), f.type, f.signature, f._env)
        self._fluent_mapping[f] = nf
        return self._env.expression_manager.FluentExp(nf, tuple(args[0].args))

    @property
    def fluent_mapping(self) -> Dict[Fluent, Fluent]:
        return self._fluent_mapping


class NegativeConditionsRemover(Transformer):
    '''Negative conditions remover class:
    this class requires a problem and offers the capability
    to transform a problem with negative conditions into one
    without negative conditions.

    This is done by substituting every fluent that appears with a Not into the conditions
    with different fluent representing  his negation.'''
    def __init__(self, problem: Problem, name: str = 'ncrm'):
        Transformer.__init__(self, problem, name)
        #NOTE no simplification are made. But it's possible to add them in key points
        self._fluent_remover = NegativeFluentRemover(self, self._env)
        #Represents the map from the new action to the old action
        self._new_to_old: Dict[Action, Action] = {}
        #represents a mapping from the action of the original problem to action of the new one.
        self._old_to_new: Dict[Action, List[Action]] = {}

    def get_rewritten_problem(self) -> Problem:
        '''Creates a problem that is a copy of the original problem
        but every ngeative fluent into action preconditions or overall
        goal is replaced by the fluent representing his negative.'''
        if self._new_problem is not None:
            return self._new_problem
        #NOTE that a different environment might be needed when multi-threading

        self._new_problem = Problem(f'{self._name}_{self._problem.name}', self._env)
        for o in self._problem.all_objects:
            self._new_problem.add_object(o)
        assert self._new_problem is not None

        name_action_map: Dict[str, Union[InstantaneousAction, DurativeAction]] = {}
        for action in self._problem.actions:
            if isinstance(action, InstantaneousAction):
                new_action = action.clone()
                new_action.name = self.get_fresh_name(action.name)
                new_action.clear_preconditions()
                for p in action.preconditions:
                    np = self._fluent_remover.remove_negative_fluents(p)
                    new_action.add_precondition(np)
                for ce in new_action.conditional_effects:
                    ce.set_condition(self._fluent_remover.remove_negative_fluents(ce.condition))
                name_action_map[action.name] = new_action
            elif isinstance(action, DurativeAction):
                new_durative_action = action.clone()
                new_durative_action.name = self.get_fresh_name(action.name)
                new_durative_action.clear_conditions()
                for i, cl in action.conditions.items():
                    for c in cl:
                        nc = self._fluent_remover.remove_negative_fluents(c)
                        new_durative_action.add_condition(i, nc)
                for t, cel in new_durative_action.conditional_effects.items():
                    for ce in cel:
                        ce.set_condition(self._fluent_remover.remove_negative_fluents(ce.condition))
                name_action_map[action.name] = new_durative_action
            else:
                raise NotImplementedError

        for t, el in self._problem.timed_effects.items():
            for e in el:
                self._new_problem._add_effect_instance(t, e.clone())

        for t, el in self._new_problem.timed_effects.items():
            for e in el:
                if e.is_conditional():
                    e.set_condition(self._fluent_remover.remove_negative_fluents(e.condition))

        for i, gl in self._problem.timed_goals.items():
            for g in gl:
                ng = self._fluent_remover.remove_negative_fluents(g)
                self._new_problem.add_timed_goal(i, ng)

        for g in self._problem.goals:
            ng = self._fluent_remover.remove_negative_fluents(g)
            self._new_problem.add_goal(ng)

        #fluent_mapping is the map between a fluent and it's negation, when the
        # negation is None it means the fluent is never found in a negation into
        # every condititon analized before; therefore it does not need to exist.
        fluent_mapping = self._fluent_remover.fluent_mapping
        for f in self._problem.fluents:
            self._new_problem.add_fluent(f)
            fneg = fluent_mapping.get(f, None)
            if fneg is not None:
                self._new_problem.add_fluent(fneg)

        for fl, v in self._problem.initial_values.items():
            fneg = fluent_mapping.get(fl.fluent(), None)
            self._new_problem.set_initial_value(fl, v)
            if fneg is not None:
                if v.bool_constant_value():
                    self._new_problem.set_initial_value(self._env.expression_manager.FluentExp(fneg,
                    tuple(fl.args)), self._env.expression_manager.FALSE())
                else:
                    self._new_problem.set_initial_value(self._env.expression_manager.FluentExp(fneg,
                    tuple(fl.args)), self._env.expression_manager.TRUE())

        for action in self._problem.actions:
            if isinstance(action, InstantaneousAction):
                new_action = name_action_map[action.name]
                new_effects: List[Effect] = []
                for e in new_action.effects:
                    fl, v = e.fluent, e.value
                    fneg = fluent_mapping.get(fl.fluent(), None)
                    if fneg is not None:
                        simplified_not_v = self._simplifier.simplify(self._env.expression_manager.Not(v))
                        new_effects.append(Effect(self._env.expression_manager.FluentExp(fneg, tuple(fl.args)),
                                                simplified_not_v, e.condition, e.kind))
                for ne in new_effects:
                    new_action._add_effect_instance(ne)
                self._new_problem.add_action(new_action)
                self._old_to_new[action] = [new_action]
                self._new_to_old[new_action] = action
            elif isinstance(action, DurativeAction):
                new_durative_action = name_action_map[action.name]
                new_durative_action.set_duration_constraint(action.duration)

                for t, el in new_durative_action.effects.items():
                    for e in el:
                        fl, v = e.fluent, e.value
                        fneg = fluent_mapping.get(fl.fluent(), None)
                        if fneg is not None:
                            simplified_not_v = self._simplifier.simplify(self._env.expression_manager.Not(v))
                            new_durative_action._add_effect_instance(t, Effect(self._env.expression_manager.FluentExp(fneg, tuple(fl.args)),
                                                            simplified_not_v, e.condition, e.kind))
                self._new_problem.add_action(new_durative_action)
                self._old_to_new[action] = [new_durative_action]
                self._new_to_old[new_durative_action] = action
            else:
                raise NotImplementedError

        for t, el in self._new_problem.timed_effects.items():
            for e in el:
                fl, v = e.fluent, e.value
                fneg = fluent_mapping.get(fl.fluent(), None)
                if fneg is not None:
                    simplified_not_v = self._simplifier.simplify(self._env.expression_manager.Not(v))
                    self._new_problem._add_effect_instance(t, Effect(self._env.expression_manager.FluentExp(fneg, tuple(fl.args)),
                                                    simplified_not_v, e.condition, e.kind))

        return self._new_problem

    def get_original_action(self, action: Action) -> Action:
        '''After the method get_rewritten_problem is called, this function maps
        the actions of the transformed problem into the actions of the original problem.'''
        return self._new_to_old[action]

    def get_transformed_actions(self, action: Action) -> List[Action]:
        '''After the method get_rewritten_problem is called, this function maps
        the actions of the original problem into the actions of the transformed problem.'''
        return self._old_to_new[action]



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
import unified_planning.solvers

from unified_planning.shortcuts import *
from unified_planning.test import TestCase, skipIfNoOneshotPlannerForProblemKind, skipIfNoPlanValidatorForProblemKind, skipIfSolverNotAvailable
from unified_planning.test.examples import get_example_problems
from unified_planning.model.problem_kind import basic_classical_kind, full_classical_kind, hierarchical_kind


class TestTarskiGrounder(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()


    @skipIfNoOneshotPlannerForProblemKind(full_classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    @skipIfSolverNotAvailable('tarski_grounder')
    def test_robot_loader(self):
        problem, plan = self.problems['robot_loader']
        with Grounder(name='tarski_grounder') as grounder:
            grounded_problem, rewrite_back_plan_function = grounder.ground(problem)
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = rewrite_back_plan_function(grounded_plan)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))


    @skipIfNoOneshotPlannerForProblemKind(full_classical_kind)
    @skipIfNoPlanValidatorForProblemKind(full_classical_kind)
    @skipIfSolverNotAvailable('tarski_grounder')
    def test_robot_locations_connected_without_battery(self):
        problem, plan = self.problems['robot_locations_connected_without_battery']
        with Grounder(name='tarski_grounder') as grounder:
            grounded_problem, rewrite_back_plan_function = grounder.ground(problem)
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = rewrite_back_plan_function(grounded_plan)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))


    @skipIfNoOneshotPlannerForProblemKind(basic_classical_kind.union(hierarchical_kind))
    @skipIfNoPlanValidatorForProblemKind(basic_classical_kind.union(hierarchical_kind))
    @skipIfSolverNotAvailable('tarski_grounder')
    def test_hierarchical_blocks_world(self):
        problem, plan = self.problems['hierarchical_blocks_world']
        with Grounder(name='tarski_grounder') as grounder:
            grounded_problem, rewrite_back_plan_function = grounder.ground(problem)
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = rewrite_back_plan_function(grounded_plan)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))


    @skipIfNoOneshotPlannerForProblemKind(basic_classical_kind)
    @skipIfSolverNotAvailable('tarski_grounder')
    def test_tarski_grounder_mockup_problem(self):
        problem = Problem('mockup')
        Location = UserType('Location')
        at = Fluent('at', BoolType(), position=Location)
        at_l1 = Fluent('at_l1')
        at_l2 = Fluent('at_l2')
        l1 = Object('l1', Location)
        l2 = Object('l2', Location)
        move_to = InstantaneousAction('move_to', l_to=Location)
        l_to = move_to.parameter('l_to')
        move_to.add_effect(at(l_to), True)
        move_to.add_effect(at_l1, False, Equals(l_to, l2))
        move_to_l2 = InstantaneousAction('move_to_l2')
        move_to_l2.add_effect(at_l2, True)
        problem.add_fluent(at, default_initial_value=False)
        problem.add_fluent(at_l1, default_initial_value=True)
        problem.add_fluent(at_l2, default_initial_value=False)
        problem.add_object(l1)
        problem.add_object(l2)
        problem.add_action(move_to)
        problem.add_action(move_to_l2)
        problem.add_goal(at(l1))
        problem.add_goal(at(l2))
        problem.add_goal(at_l2)

        with Grounder(name='tarski_grounder') as grounder:
            grounded_problem, rewrite_back_plan_function = grounder.ground(problem)
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = rewrite_back_plan_function(grounded_plan)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(problem_kind=problem.kind) as pv:
                    self.assertTrue(pv.validate(problem, plan))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from fractions import Fraction
from itertools import count
import re
import sys
from unicodedata import name
import unified_planning as up
import unified_planning.walkers as walkers
from unified_planning.model.types import _UserType, _IntType, _RealType
from typing import Dict, IO, Optional, cast
from io import StringIO


class ConverterToPythonString(walkers.DagWalker):
    '''Expression converter to a Python string.'''

    def __init__(self, env: 'up.environment.Environment', names_mapping: Dict[str, str]):
        walkers.DagWalker.__init__(self)
        self._names_mapping = names_mapping

    def convert(self, expression):
        '''Converts the given expression to a python command.'''
        if expression is None:
            return 'None'
        return self.walk(expression)

    def walk_exists(self, expression, args):
        assert len(args) == 1
        vars_string_list = [f'up.model.Variable("{v.name}", {_print_python_type(v.type, self._names_mapping)})' for v in expression.variables()]
        return f'emgr.Exists({args[0]}, {", ".join(vars_string_list)})'

    def walk_forall(self, expression, args):
        assert len(args) == 1
        vars_string_list = [f'up.model.Variable("{v.name}", {_print_python_type(v.type, self._names_mapping)})' for v in expression.variables()]
        return f'emgr.Forall({args[0]}, {", ".join(vars_string_list)})'

    def walk_variable_exp(self, expression, args):
        assert len(args) == 0
        v = expression.variable()
        return f'emgr.VariableExp(up.model.Variable("{v.name}", {_print_python_type(v.type, self._names_mapping)}))'

    def walk_and(self, expression, args):
        assert len(args) > 1
        return f'emgr.And({", ".join(args)})'

    def walk_or(self, expression, args):
        assert len(args) > 1
        return f'emgr.Or({", ".join(args)})'

    def walk_not(self, expression, args):
        assert len(args) == 1
        return f'emgr.Not({args[0]})'

    def walk_implies(self, expression, args):
        assert len(args) == 2
        return f'emgr.Implies({args[0]}, {args[1]})'

    def walk_iff(self, expression, args):
        assert len(args) == 2
        return f'emgr.Iff({args[0]}, {args[1]})'

    def walk_fluent_exp(self, expression, args):
        fluent = expression.fluent()
        if args:
            return f'{_get_mangled_name(f"fluent_{fluent.name}", self._names_mapping)}({", ".join(args)})'
        return f'emgr.FluentExp({_get_mangled_name(f"fluent_{fluent.name}", self._names_mapping)})'

    def walk_param_exp(self, expression, args):
        assert len(args) == 0
        p = expression.parameter()
        return f'emgr.ParameterExp(up.model.Parameter("{p.name}", {_print_python_type(p.type, self._names_mapping)}))'

    def walk_object_exp(self, expression, args):
        assert len(args) == 0
        o = expression.object()
        return f'emgr.ObjectExp({_get_mangled_name(f"object_{o.name}", self._names_mapping)})'

    def walk_bool_constant(self, expression, args):
        assert len(args) == 0
        if expression.bool_constant_value():
            return 'emgr.TRUE()'
        else:
            return 'emgr.FALSE()'

    def walk_real_constant(self, expression, args):
        assert len(args) == 0
        return f'emgr.Real(Fraction({str(expression.real_constant_value().numerator)}, {str(expression.real_constant_value().denominator)}))'

    def walk_int_constant(self, expression, args):
        assert len(args) == 0
        return f'emgr.Int({str(expression.constant_value())})'

    def walk_plus(self, expression, args):
        assert len(args) > 1
        return f'emgr.Plus({", ".join(args)})'

    def walk_minus(self, expression, args):
        assert len(args) == 2
        return f'emgr.Minus({args[0]}, {args[1]})'

    def walk_times(self, expression, args):
        assert len(args) > 1
        return f'emgr.Times({", ".join(args)})'

    def walk_div(self, expression, args):
        assert len(args) == 2
        return f'emgr.Div({args[0]}, {args[1]})'

    def walk_le(self, expression, args):
        assert len(args) == 2
        return f'emgr.LE({args[0]}, {args[1]})'

    def walk_lt(self, expression, args):
        assert len(args) == 2
        return f'emgr.LT({args[0]}, {args[1]})'

    def walk_equals(self, expression, args):
        assert len(args) == 2
        return f'emgr.Equals({args[0]}, {args[1]})'


class PythonWriter:
    '''This class can be used to write a Problem in PDDL.'''

    def __init__(self, problem: 'up.model.Problem'):
        self.problem = problem

    def _write_problem_code(self, out: IO[str]):

        names_mapping: Dict[str, str] = {}
        # Initialize names_mapping with valid names
        for type in self.problem.user_types: # define user_types
            utype = cast(_UserType, type)
            type_var_name = f'type_{utype.name}'
            if type_var_name.isidentifier():
                names_mapping[type_var_name] = type_var_name
        for fluent in self.problem.fluents:
            fluent_var_name = f'fluent_{fluent.name}'
            if fluent_var_name.isidentifier():
                names_mapping[fluent_var_name] = fluent_var_name
        for object in self.problem.all_objects:
            object_var_name = f'object_{object.name}'
            if object_var_name.isidentifier():
                names_mapping[object_var_name] = object_var_name
        for action in self.problem.actions:
            action_var_name = f'act_{action.name}'
            if action_var_name.isidentifier():
                names_mapping[action_var_name] = action_var_name

        converter = ConverterToPythonString(self.problem.env, names_mapping)
        out.write('from fractions import Fraction\n')
        out.write('from collections import OrderedDict\n')
        out.write('import unified_planning as up\n')
        out.write('env = up.environment.get_env()\n')
        out.write('emgr = env.expression_manager\n')
        out.write('tm = env.type_manager\n')

        for type in self.problem.user_types: # define user_types
            utype = cast(_UserType, type)
            if utype.father is None:
                out.write(f'{_get_mangled_name(f"type_{utype.name}", names_mapping)} = tm.UserType("{utype.name}")\n')
            else:
                out.write(f'{_get_mangled_name(f"type_{utype.name}", names_mapping)} = tm.UserType("{utype.name}", {_get_mangled_name(f"type_{cast(_UserType, utype.father).name}", names_mapping)})\n')

        for f in self.problem.fluents: # define fluents
            params = ', '.join(f'("{p.name}", {_print_python_type(p.type, names_mapping)})' for p in f.signature)
            params = f'OrderedDict([{params}])'
            out.write(f'{_get_mangled_name(f"fluent_{f.name}", names_mapping)} = up.model.Fluent("{f.name}", {_print_python_type(f.type, names_mapping)}, _signature={params})\n')

        for o in self.problem.all_objects: # define objects
            out.write(f'{_get_mangled_name(f"object_{o.name}", names_mapping)} = up.model.Object("{o.name}", {_get_mangled_name(f"type_{cast(_UserType, o.type).name}", names_mapping)})\n')

        out.write('problem_initial_defaults = {}\n') # define initial_defaults
        for type, exp in self.problem.initial_defaults.items():
            out.write(f'problem_initial_defaults[{_print_python_type(type, names_mapping)}] = {converter.convert(exp)}\n')
        out.write(f'problem = up.model.Problem("{self.problem.name}", env, initial_defaults=problem_initial_defaults)\n')

        for o in self.problem.all_objects: # add objects to the problem
            out.write(f'problem.add_object({_get_mangled_name(f"object_{o.name}", names_mapping)})\n')

        for f in self.problem.fluents: # add fluents to the problem, with their fluents_default, if they have one
            default = self.problem.fluents_defaults.get(f, None)
            out.write(f'problem.add_fluent({_get_mangled_name(f"fluent_{f.name}", names_mapping)}')
            if default is not None: # the fluent has a default value
                out.write(f', default_initial_value={converter.convert(default)}')
            out.write(')\n')

        for a in self.problem.actions: # define actions and add them to the problem
            if isinstance(a, up.model.InstantaneousAction):
                params = ', '.join(f'("{p.name}", {_print_python_type(p.type, names_mapping)})' for p in a.parameters)
                params = f'OrderedDict([{params}])'
                out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)} = up.model.InstantaneousAction("{a.name}", _parameters={params})\n')
                for p in a.preconditions:
                    out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_precondition({converter.convert(p)})\n')
                for e in a.effects:
                    if e.is_increase():
                        out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_increase_effect(fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')
                    elif e.is_decrease():
                        out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_decrease_effect(fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')
                    else:
                        out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_effect(fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')
            elif isinstance(a, up.model.DurativeAction):
                out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)} = up.model.DurativeAction("{a.name}"')
                for ap in a.parameters:
                    out.write(f', {ap.name}={_print_python_type(ap.type, names_mapping)}')
                out.write(')\n')
                for ap in a.parameters:
                    out.write(f'parameter_{ap.name} = {_get_mangled_name(f"act_{a.name}", names_mapping)}.parameter("{ap.name}")\n')
                out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.set_duration_constraint({_convert_interval_duration(a.duration, converter)})\n')
                for i, cl in a.conditions.items():
                    for c in cl:
                        out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_condition({_convert_interval(i)}, {converter.convert(c)})\n')
                for t, el in a.effects.items():
                    for e in el:
                        if e.is_increase():
                            out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_increase_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')
                        elif e.is_decrease():
                            out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_decrease_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')
                        else:
                            out.write(f'{_get_mangled_name(f"act_{a.name}", names_mapping)}.add_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')
            else:
                raise NotImplementedError
            out.write(f'problem.add_action({_get_mangled_name(f"act_{a.name}", names_mapping)})\n')

        for f_exp, v_exp in self.problem.explicit_initial_values.items(): # add only previously added initial values
            out.write(f'problem.set_initial_value({converter.convert(f_exp)}, {converter.convert(v_exp)})\n')

        for t, el in self.problem.timed_effects.items(): # add timed effects
            for e in el:
                if e.is_increase():
                    out.write(f'problem.add_increase_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')
                elif e.is_decrease():
                    out.write(f'problem.add_decrease_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')
                else:
                    out.write(f'problem.add_timed_effect(timing={_convert_timing(t)}, fluent={converter.convert(e.fluent)}, value={converter.convert(e.value)}, condition={converter.convert(e.condition)})\n')

        for i, gl in self.problem.timed_goals.items(): # add timed goals
            for g in gl:
                out.write(f'problem.add_timed_goal(interval={_convert_interval(i)}, goal={converter.convert(g)})\n')

        for g in self.problem.goals: # add goals
            out.write(f'problem.add_goal(goal={converter.convert(g)})\n')

        for qm in self.problem.quality_metrics: # adding metrics
            if isinstance(qm, up.model.metrics.MinimizeActionCosts):
                out.write('costs = {}\n')
                for a, c in qm.costs.items():
                    out.write(f'costs[act_{a.name}] = {converter.convert(c)}\n')
            out.write('problem.add_quality_metric(')
            if isinstance(qm, up.model.metrics.MinimizeActionCosts):
                out.write(f'up.model.metrics.MinimizeActionCosts(costs, {qm.default})')
            elif isinstance(qm, up.model.metrics.MinimizeSequentialPlanLength):
                out.write('up.model.metrics.MinimizeSequentialPlanLength()')
            elif isinstance(qm, up.model.metrics.MinimizeMakespan):
                out.write('up.model.metrics.MinimizeMakespan()')
            elif isinstance(qm, up.model.metrics.MinimizeExpressionOnFinalState):
                out.write(f'up.model.metrics.MinimizeExpressionOnFinalState({converter.convert(qm.expression)})')
            elif isinstance(qm, up.model.metrics.MaximizeExpressionOnFinalState):
                out.write(f'up.model.metrics.MaximizeExpressionOnFinalState({converter.convert(qm.expression)})')
            else:
                raise NotImplementedError
            out.write(')\n')

    def print_problem_python_commands(self):
        '''Prints the string representing all the necessary commands to recreate the problem.'''
        self._write_problem_code(sys.stdout)

    def write_problem_code(self) -> str:
        '''Returns the string representing all the necessary commands to recreate the problem.'''
        out = StringIO()
        self._write_problem_code(out)
        return out.getvalue()

    def write_problem_code_to_file(self, filename: str):
        '''Dumps to file the PDDL domain.'''
        with open(filename, 'w') as f:
            self._write_problem_code(f)


def _get_mangled_name(original_name: str, names_mapping: Dict[str, str]) -> str:
    '''Important note: This method updates the names_mapping '''
    regex = re.compile('^[a-zA-Z_]+.*')
    assert re.match(regex, original_name) is not None
    new_name: Optional[str] = names_mapping.get(original_name, None)
    if new_name is None: # The name is not in the map, so it must be added
        if not original_name.isidentifier(): # Make the name a python identifier
            new_name = re.sub('[^0-9a-zA-Z_]', '_', f'{original_name}')
        else:
            new_name = original_name
        test_name = new_name # Init values
        count = 0
        while test_name in names_mapping.values(): # Loop until we find a fresh name
            test_name = f'{new_name}_{str(count)}'
            count += 1
        new_name = test_name
        names_mapping[original_name] = new_name # Once a fresh valid name is found, update the map.
    assert new_name is not None
    assert new_name.isidentifier()
    return new_name

def _print_python_type(type: 'up.model.types.Type', names_mapping: Dict[str, str]) -> str:
    '''This method takes a type and returns how to use it from command line.
    For the user_types it assumes they have already been created and just refers to them.'''
    if type.is_user_type():
        return f'{_get_mangled_name(f"type_{cast(_UserType, type).name}", names_mapping)}'
    elif type.is_bool_type():
        return 'tm.BoolType()'
    elif type.is_int_type():
        itype = cast(_IntType, type)
        return f'tm.IntType({str(itype.lower_bound)}, {str(itype.upper_bound)})'
    elif type.is_real_type():
        rtype = cast(_RealType, type)
        return f'tm.RealType({str(rtype.lower_bound)}, {str(rtype.upper_bound)})'
    else:
        raise NotImplementedError

def _convert_timing(timing: up.model.Timing) -> str:
    delay: str = f'{str(timing.delay)}'
    if isinstance(timing.delay, Fraction):
        delay = f'Fraction({str(timing.delay.numerator)}, {str(timing.delay.denominator)})'
    if timing.is_from_start():
        return f'up.model.StartTiming({delay})'
    else:
        return f'up.model.EndTiming({delay})'

def _convert_interval(interval: up.model.TimeInterval) -> str:
    interval_feature: str = 'up.model.ClosedTimeInterval'
    if interval.is_left_open() and interval.is_right_open():
        interval_feature = 'up.model.OpenTimeInterval'
    elif interval.is_left_open() and not interval.is_right_open():
        interval_feature = 'up.model.LeftOpenTimeInterval'
    elif not interval.is_left_open() and interval.is_right_open():
        interval_feature = 'up.model.RightOpenTimeInterval'
    return f'{interval_feature}({_convert_timing(interval.lower)}, {_convert_timing(interval.upper)})'

def _convert_interval_duration(interval: up.model.DurationInterval, converter: ConverterToPythonString) -> str:
    interval_feature: str = 'up.model.ClosedDurationInterval'
    if interval.is_left_open() and interval.is_right_open():
        interval_feature = 'up.model.OpenDurationInterval'
    elif interval.is_left_open() and not interval.is_right_open():
        interval_feature = 'up.model.LeftOpenDurationInterval'
    elif not interval.is_left_open() and interval.is_right_open():
        interval_feature = 'up.model.RightOpenDurationInterval'
    return f'{interval_feature}({converter.convert(interval.lower)}, {converter.convert(interval.upper)})'



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from unified_planning.model.action import Action, InstantaneousAction, DurativeAction
from unified_planning.model.effect import Effect, EffectKind
from unified_planning.model.expression import BoolExpression, Expression, ExpressionManager
from unified_planning.model.fnode import FNode
from unified_planning.model.fluent import Fluent
from unified_planning.model.object import Object
from unified_planning.model.operators import OperatorKind
from unified_planning.model.parameter import Parameter
from unified_planning.model.problem import Problem
from unified_planning.model.problem_kind import ProblemKind
from unified_planning.model.timing import TimepointKind, Timing, StartTiming, EndTiming, GlobalStartTiming, GlobalEndTiming, DurationInterval, ClosedDurationInterval
from unified_planning.model.timing import FixedDuration, OpenDurationInterval, LeftOpenDurationInterval, RightOpenDurationInterval
from unified_planning.model.timing import TimeInterval, TimePointInterval, ClosedTimeInterval, OpenTimeInterval, LeftOpenTimeInterval, RightOpenTimeInterval
from unified_planning.model.types import Type, TypeManager
from unified_planning.model.variable import Variable, FreeVarsOracle
import unified_planning.model.metrics

__all__ = [ 'Action', 'InstantaneousAction', 'DurativeAction',
            'Effect', 'EffectKind',
            'BoolExpression', 'Expression', 'ExpressionManager',
            'FNode',
            'Fluent',
            'Object',
            'OperatorKind',
            'Parameter',
            'Problem',
            'ProblemKind',
            'TimepointKind', 'Timing', 'StartTiming', 'EndTiming', 'GlobalStartTiming', 'GlobalEndTiming',
            'DurationInterval', 'ClosedDurationInterval', 'FixedDuration', 'OpenDurationInterval', 'LeftOpenDurationInterval', 'RightOpenDurationInterval',
            'TimeInterval', 'TimePointInterval', 'ClosedTimeInterval', 'OpenTimeInterval', 'LeftOpenTimeInterval', 'RightOpenTimeInterval',
            'Type', 'TypeManager',
            'Variable', 'FreeVarsOracle'
            ]



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from unified_planning.solvers.solver import Solver, OptimalityGuarantee
from unified_planning.solvers.factory import Factory
from unified_planning.solvers.grounder import Grounder
from unified_planning.solvers.parallel import Parallel
from unified_planning.solvers.pddl_solver import PDDLSolver
from unified_planning.solvers.plan_validator import SequentialPlanValidator
from unified_planning.solvers.results import LogMessage, PlanGenerationResult, LogLevel, PlanGenerationResultStatus

__all__ = [ 'Factory',
            'Grounder',
            'Parallel',
            'PDDLSolver',
            'SequentialPlanValidator',
            'Solver', 'OptimalityGuarantee',
            'LogMessage', 'PlanGenerationResult', 'LogLevel', 'PlanGenerationResultStatus'
        ]



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.walkers import Substituter
from unified_planning.environment import get_env
from unified_planning.exceptions import UPTypeError

class TestSubstituter(TestCase):
    def setUp(self):
        TestCase.setUp(self)

    def test_id_walker(self):
        s = Substituter(get_env())
        #small test on already-done expressions to check the id-dagwalker
        x = FluentExp(Fluent('x'))
        y = FluentExp(Fluent('y', IntType()))
        t = Bool(True)
        f = Bool(False)
        subs = {y: 3}
        # ((25/5)*30*2*2) - (20*5) (500) == (25*4*10) / 2 (500)
        e1 = Equals(Minus(Times([Div(25, 5), 30, 2, 2]), Times(20, 5)), Div(Times(25, 4, 10) ,2))
        r1 = s.substitute(e1, subs)
        self.assertEqual(r1, e1)
        # T => !x
        e2 = Implies(e1, Not(x))
        r2 = s.substitute(e2)
        self.assertEqual(r2, e2, subs)
        # !x || (T => x)
        e3 = Or(e2, Implies(e1, x))
        r3 = s.substitute(e3, subs)
        self.assertEqual(r3, e3)

    def test_substitution(self):
        s = Substituter(get_env())
        xfluent = Fluent('x', IntType())
        x = FluentExp(xfluent)
        subst = {}
        subst[x] = Int(5)
        e1 = Plus(x, 1)
        s1 = s.substitute(e1, subst)
        self.assertEqual(s1, Plus(5, 1))
        #Testing that (a & b) with sbust = {a <- c, (c & b) <- d, (a & b) <- c} is c
        a = Fluent('a')
        b = Fluent('b')
        c = FluentExp(Fluent('c'))
        d = Fluent('d')
        subst = {}
        subst[a] = c
        subst[And(c,b)] = d
        subst[And(a,b)] = c
        e2 = And(a, b)
        s2 = s.substitute(e2, subst)
        self.assertEqual(s2, c)

        with self.assertRaises(UPTypeError):
            subst = {}
            subst[a] = c
            subst[And(c,b)] = d
            subst[And(a,b)] = Int(5)
            e3 = And(a, b)
            s3 = s.substitute(e3, subst)

        subst = {}
        subst[a] = c
        subst[And(c,b)] = d
        e4 = And(a, b, And(c, b))
        s4 = s.substitute(e4, subst)
        self.assertEqual(s4, And(c,b,d))

        subst = {}
        subst[a] = c
        subst[c] = d
        e5 = And(a, c, And(a, c))
        s5 = s.substitute(e5, subst)
        self.assertEqual(s5, And(c, d, And(c, d)))

        with self.assertRaises(UPTypeError):
            subst = {}
            subst[a] = c
            subst[And(c,b)] = Plus(1, 2)
            e6 = Int(1)
            s.substitute(e6, subst)


if __name__ == "__main__":
    main()



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""FNode are the building blocks of expressions."""

import unified_planning
import unified_planning.model.fluent
import collections
from unified_planning.model.operators import OperatorKind
from typing import List, Union
from fractions import Fraction

FNodeContent = collections.namedtuple("FNodeContent",
                                      ["node_type", "args", "payload"])

class FNode(object):
    __slots__ = ["_content", "_node_id"]

    def __init__(self, content: FNodeContent, node_id: int):
        self._content = content
        self._node_id = node_id
        return

    # __eq__ is left as default while __hash__ uses the node id. This
    # is because we always have shared FNodes, hence in a single
    # environment two nodes have always different ids, but in
    # different environments they can have the same id. This is not an
    # issue since, by default, equality coincides with identity.
    def __hash__(self) -> int:
        return self._node_id

    def get_nary_expression_string(self, op: str, args: List['FNode']) -> str:
        p = []
        if len(args) > 0:
            p.append('(')
            p.append(str(args[0]))
            for x in args[1:]:
                p.append(op)
                p.append(str(x))
            p.append(')')
        return ''.join(p)

    def __repr__(self) -> str:
        if self.is_bool_constant():
            return 'true' if self.is_true() else 'false'
        elif self.is_int_constant():
            return str(self.constant_value())
        elif self.is_real_constant():
            return str(self.constant_value())
        elif self.is_fluent_exp():
            return self.fluent().name + self.get_nary_expression_string(', ', self.args)
        elif self.is_parameter_exp():
            return self.parameter().name
        elif self.is_variable_exp():
            return self.variable().name
        elif self.is_object_exp():
            return self.object().name
        elif self.is_and():
            return self.get_nary_expression_string(' and ', self.args)
        elif self.is_or():
            return self.get_nary_expression_string(' or ', self.args)
        elif self.is_not():
            return f'(not {str(self.arg(0))})'
        elif self.is_implies():
            return self.get_nary_expression_string(' implies ', self.args)
        elif self.is_iff():
            return self.get_nary_expression_string(' iff ', self.args)
        elif self.is_exists():
            s = ', '.join(str(v) for v in self.variables())
            return f"Exists ({s}) {str(self.arg(0))}"
        elif self.is_forall():
            s = ', '.join(str(v) for v in self.variables())
            return f"Forall ({s}) {str(self.arg(0))}"
        elif self.is_plus():
            return self.get_nary_expression_string(' + ', self.args)
        elif self.is_minus():
            return self.get_nary_expression_string(' - ', self.args)
        elif self.is_times():
            return self.get_nary_expression_string(' * ', self.args)
        elif self.is_div():
            return self.get_nary_expression_string(' / ', self.args)
        elif self.is_le():
            return self.get_nary_expression_string(' <= ', self.args)
        elif self.is_lt():
            return self.get_nary_expression_string(' < ', self.args)
        elif self.is_equals():
            return self.get_nary_expression_string(' == ', self.args)
        else:
            raise

    @property
    def node_id(self) -> int:
        return self._node_id

    @property
    def node_type(self) -> OperatorKind:
        return self._content.node_type

    @property
    def args(self) -> List['FNode']:
        """Returns the subexpressions."""
        return self._content.args

    def arg(self, idx: int) -> 'FNode':
        """Return the given subexpression at the given position."""
        return self._content.args[idx]

    def is_constant(self) -> bool:
        """Test whether the expression is a constant."""
        return self.node_type == OperatorKind.BOOL_CONSTANT or \
            self.node_type == OperatorKind.INT_CONSTANT or \
            self.node_type == OperatorKind.REAL_CONSTANT or \
            self.node_type == OperatorKind.OBJECT_EXP

    def constant_value(self) -> Union[bool, int, Fraction]:
        """Return the value of the Constant."""
        assert self.is_constant()
        return self._content.payload

    def bool_constant_value(self) -> bool:
        """Return the bool value of the Constant."""
        assert self.is_bool_constant()
        return self._content.payload

    def int_constant_value(self) -> int:
        """Return the int value of the Constant."""
        assert self.is_int_constant()
        return self._content.payload

    def real_constant_value(self) -> Fraction:
        """Return the real value of the Constant."""
        assert self.is_real_constant()
        return self._content.payload

    def fluent(self) -> 'unified_planning.model.fluent.Fluent':
        """Return the fluent of the FluentExp."""
        assert self.is_fluent_exp()
        return self._content.payload

    def parameter(self) -> 'unified_planning.model.parameter.Parameter':
        """Return the parameter of the ParameterExp."""
        assert self.is_parameter_exp()
        return self._content.payload

    def variable(self) -> 'unified_planning.model.variable.Variable':
        """Return the variable of the VariableExp."""
        assert self.is_variable_exp()
        return self._content.payload

    def variables(self) -> List['unified_planning.model.variable.Variable']:
        """Return the variable of the Exists or Forall."""
        assert self.is_exists() or self.is_forall()
        return list(self._content.payload)

    def object(self) -> 'unified_planning.model.object.Object':
        """Return the object of the ObjectExp."""
        assert self.is_object_exp()
        return self._content.payload

    def is_bool_constant(self) -> bool:
        """Test whether the expression is a boolean constant."""
        return self.node_type == OperatorKind.BOOL_CONSTANT

    def is_int_constant(self) -> bool:
        """Test whether the expression is an integer constant."""
        return self.node_type == OperatorKind.INT_CONSTANT

    def is_real_constant(self) -> bool:
        """Test whether the expression is a real constant."""
        return self.node_type == OperatorKind.REAL_CONSTANT

    def is_true(self) -> bool:
        """Test whether the expression is the True Boolean constant."""
        return self.is_bool_constant() and self.constant_value() == True

    def is_false(self) -> bool:
        """Test whether the expression is the False Boolean constant."""
        return self.is_bool_constant() and self.constant_value() == False

    def is_and(self) -> bool:
        """Test whether the node is the And operator."""
        return self.node_type == OperatorKind.AND

    def is_or(self) -> bool:
        """Test whether the node is the Or operator."""
        return self.node_type == OperatorKind.OR

    def is_not(self) -> bool:
        """Test whether the node is the Not operator."""
        return self.node_type == OperatorKind.NOT

    def is_implies(self) -> bool:
        """Test whether the node is the Implies operator."""
        return self.node_type == OperatorKind.IMPLIES

    def is_iff(self) -> bool:
        """Test whether the node is the Iff operator."""
        return self.node_type == OperatorKind.IFF

    def is_exists(self) -> bool:
        """Test whether the node is the Exists operator."""
        return self.node_type == OperatorKind.EXISTS

    def is_forall(self) -> bool:
        """Test whether the node is the Forall operator."""
        return self.node_type == OperatorKind.FORALL

    def is_fluent_exp(self) -> bool:
        """Test whether the node is a fluent."""
        return self.node_type == OperatorKind.FLUENT_EXP

    def is_parameter_exp(self) -> bool:
        """Test whether the node is an action parameter."""
        return self.node_type == OperatorKind.PARAM_EXP

    def is_variable_exp(self) -> bool:
        """Test whether the node is a variable."""
        return self.node_type == OperatorKind.VARIABLE_EXP

    def is_object_exp(self) -> bool:
        """Test whether the node is an action object."""
        return self.node_type == OperatorKind.OBJECT_EXP

    def is_plus(self) -> bool:
        """Test whether the node is the Plus operator."""
        return self.node_type == OperatorKind.PLUS

    def is_minus(self) -> bool:
        """Test whether the node is the Minus operator."""
        return self.node_type == OperatorKind.MINUS

    def is_times(self) -> bool:
        """Test whether the node is the Times operator."""
        return self.node_type == OperatorKind.TIMES

    def is_div(self) -> bool:
        """Test whether the node is the Div operator."""
        return self.node_type == OperatorKind.DIV

    def is_equals(self) -> bool:
        """Test whether the node is the Equals operator."""
        return self.node_type == OperatorKind.EQUALS

    def is_le(self) -> bool:
        """Test whether the node is the LE operator."""
        return self.node_type == OperatorKind.LE

    def is_lt(self) -> bool:
        """Test whether the node is the LT operator."""
        return self.node_type == OperatorKind.LT
